# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
Native contacts analysis --- :mod:`MDAnalysis.analysis.contacts`
================================================================

This module contains classes to analyze native contacts *Q* over a
trajectory. Native contacts of a conformation are contacts that exist
in a reference structure and in the conformation. Contacts in the
reference structure are always defined as being closer than a distance
`radius`. The fraction of native contacts for a conformation can be
calculated in different ways. This module supports 3 different metrics
listed below, as well as custom metrics.

1. *Hard Cut*: To count as a contact the atoms *i* and *j* have to be at least
   as close as in the reference structure.

2. *Soft Cut*: The atom pair *i* and *j* is assigned based on a soft potential
   that is 1 if the distance is 0, 1/2 if the distance is the same as in
   the reference and 0 for large distances. For the exact definition of the
   potential and parameters have a look at function :func:`soft_cut_q`.

3. *Radius Cut*: To count as a contact the atoms *i* and *j* cannot be further
   apart than some distance `radius`.

The "fraction of native contacts" *Q(t)* is a number between 0 and 1 and
calculated as the total number of native contacts for a given time frame
divided by the total number of contacts in the reference structure.


Examples for contact analysis
-----------------------------

One-dimensional contact analysis
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

As an example we analyze the opening ("unzipping") of salt bridges
when the AdK enzyme opens up; this is one of the example trajectories
in MDAnalysis. ::

    import numpy as np
    import matplotlib.pyplot as plt
    import MDAnalysis as mda
    from MDAnalysis.analysis import contacts
    from MDAnalysis.tests.datafiles import PSF,DCD
    # example trajectory (transition of AdK from closed to open)
    u = mda.Universe(PSF,DCD)
    # crude definition of salt bridges as contacts between NH/NZ in ARG/LYS and
    # OE*/OD* in ASP/GLU. You might want to think a little bit harder about the
    # problem before using this for real work.
    sel_basic = "(resname ARG LYS) and (name NH* NZ)"
    sel_acidic = "(resname ASP GLU) and (name OE* OD*)"
    # reference groups (first frame of the trajectory, but you could also use a
    # separate PDB, eg crystal structure)
    acidic = u.select_atoms(sel_acidic)
    basic = u.select_atoms(sel_basic)
    # set up analysis of native contacts ("salt bridges"); salt bridges have a
    # distance <6 A
    ca1 = contacts.Contacts(u, select=(sel_acidic, sel_basic),
                            refgroup=(acidic, basic), radius=6.0)
    # iterate through trajectory and perform analysis of "native contacts" Q
    ca1.run()
    # print number of averave contacts
    average_contacts = np.mean(ca1.results.timeseries[:, 1])
    print('average contacts = {}'.format(average_contacts))
    # plot time series q(t)
    fig, ax = plt.subplots()
    ax.plot(ca1.results.timeseries[:, 0], ca1.results.timeseries[:, 1])
    ax.set(xlabel='frame', ylabel='fraction of native contacts',
           title='Native Contacts, average = {:.2f}'.format(average_contacts))
    fig.show()


The first graph shows that when AdK opens, about 20% of the salt
bridges that existed in the closed state disappear when the enzyme
opens. They open in a step-wise fashion (made more clear by the movie
`AdK_zipper_cartoon.avi`_).

.. _`AdK_zipper_cartoon.avi`:
   http://www.ncbi.nlm.nih.gov/pmc/articles/PMC2803350/bin/NIHMS150766-supplement-03.avi

.. rubric:: Notes

Suggested cutoff distances for different simulations

* For all-atom simulations, cutoff = 4.5 Å
* For coarse-grained simulations, cutoff = 6.0 Å


Two-dimensional contact analysis (q1-q2)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Analyze a single DIMS transition of AdK between its closed and open
conformation and plot the trajectory projected on q1-q2
:footcite:p:`Franklin2007` ::


    import MDAnalysis as mda
    from MDAnalysis.analysis import contacts
    from MDAnalysisTests.datafiles import PSF, DCD
    u = mda.Universe(PSF, DCD)
    q1q2 = contacts.q1q2(u, 'name CA', radius=8)
    q1q2.run()

    f, ax = plt.subplots(1, 2, figsize=plt.figaspect(0.5))
    ax[0].plot(q1q2.results.timeseries[:, 0], q1q2.results.timeseries[:, 1],
               label='q1')
    ax[0].plot(q1q2.results.timeseries[:, 0], q1q2.results.timeseries[:, 2],
               label='q2')
    ax[0].legend(loc='best')
    ax[1].plot(q1q2.results.timeseries[:, 1],
               q1q2.results.timeseries[:, 2], '.-')
    f.show()

Compare the resulting pathway to the `MinActionPath result for AdK`_
:footcite:p:`Franklin2007`.

.. _MinActionPath result for AdK:
   http://lorentz.dynstr.pasteur.fr/joel/adenylate.php


Writing your own contact analysis
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The :class:`Contacts` class has been designed to be extensible for your own
analysis. As an example we will analyze when the acidic and basic groups of AdK
are in contact which each other; this means that at least one of the contacts
formed in the reference is closer than 2.5 Å.

For this we define a new function to determine if any contact is closer than
2.5 Å; this function must implement the API prescribed by :class:`Contacts`::

    def is_any_closer(r, r0, dist=2.5):
        return np.any(r < dist)

The first two parameters `r` and `r0` are provided by :class:`Contacts` when it
calls :func:`is_any_closer` while the others can be passed as keyword args
using the `kwargs` parameter in :class:`Contacts`.

Next we are creating an instance of the :class:`Contacts` class and use the
:func:`is_any_closer` function as an argument to `method` and run the analysis::

    # crude definition of salt bridges as contacts between NH/NZ in ARG/LYS and
    # OE*/OD* in ASP/GLU. You might want to think a little bit harder about the
    # problem before using this for real work.
    sel_basic = "(resname ARG LYS) and (name NH* NZ)"
    sel_acidic = "(resname ASP GLU) and (name OE* OD*)"

    # reference groups (first frame of the trajectory, but you could also use a
    # separate PDB, eg crystal structure)
    acidic = u.select_atoms(sel_acidic)
    basic = u.select_atoms(sel_basic)

    nc = contacts.Contacts(u, select=(sel_acidic, sel_basic),
                           method=is_any_closer,
                           refgroup=(acidic, basic), kwargs={'dist': 2.5})
    nc.run()

    bound = nc.results.timeseries[:, 1]
    frames = nc.results.timeseries[:, 0]

    f, ax = plt.subplots()

    ax.plot(frames, bound, '.')
    ax.set(xlabel='frame', ylabel='is Bound',
           ylim=(-0.1, 1.1))

    f.show()


Functions
---------

.. autofunction:: hard_cut_q
.. autofunction:: soft_cut_q
.. autofunction:: radius_cut_q
.. autofunction:: contact_matrix
.. autofunction:: q1q2

Classes
-------

.. autoclass:: Contacts
   :members:

.. rubric:: References
.. footbibliography::

"""
import os
import errno
import warnings
import bz2
import functools

import numpy as np

import logging

import MDAnalysis
import MDAnalysis.lib.distances
from MDAnalysis.lib.util import openany
from MDAnalysis.analysis.distances import distance_array
from MDAnalysis.core.groups import AtomGroup, UpdatingAtomGroup
from .base import AnalysisBase, ResultsGroup

logger = logging.getLogger("MDAnalysis.analysis.contacts")


def soft_cut_q(r, r0, beta=5.0, lambda_constant=1.8):
    r"""Calculate fraction of native contacts *Q* for a soft cut off

    The native contact function is defined as :footcite:p:`Best2013`

    .. math::

        Q(r, r_0) = \frac{1}{1 + e^{\beta (r - \lambda r_0)}}

    Reasonable values for different simulation types are

    - *All Atom*: `lambda_constant = 1.8` (unitless)
    - *Coarse Grained*: `lambda_constant = 1.5` (unitless)

    Parameters
    ----------
    r: array
      Contact distances at time t
    r0: array
      Contact distances at time t=0, reference distances
    beta: float (default 5.0 Angstrom)
      Softness of the switching function
    lambda_constant: float (default 1.8, unitless)
      Reference distance tolerance

    Returns
    -------
    Q : float
      fraction of native contacts
    """
    r = np.asarray(r)
    r0 = np.asarray(r0)
    result = 1 / (1 + np.exp(beta * (r - lambda_constant * r0)))

    return result.sum() / len(r0)


def hard_cut_q(r, cutoff):
    """Calculate fraction of native contacts *Q* for a hard cut off.

    The cutoff can either be a float or a :class:`~numpy.ndarray` of the same
    shape as `r`.

    Parameters
    ----------
    r : ndarray
        distance matrix
    cutoff : ndarray | float
        cut off value to count distances. Can either be a float of a ndarray of
        the same size as distances

    Returns
    -------
    Q : float
        fraction of contacts

    """
    r = np.asarray(r)
    cutoff = np.asarray(cutoff)
    y = r <= cutoff
    return y.sum() / r.size


def radius_cut_q(r, r0, radius):
    """calculate native contacts *Q* based on the single distance radius.

    Parameters
    ----------
    r : ndarray
        distance array between atoms
    r0 : ndarray
        unused to fullfill :class:`Contacts` API
    radius : float
        Distance between atoms at which a contact is formed

    Returns
    -------
    Q : float
        fraction of contacts

    """
    return hard_cut_q(r, radius)


def contact_matrix(d, radius, out=None):
    """calculate contacts from distance matrix

    Parameters
    ----------
    d : array-like
        distance matrix
    radius : float
        distance below which a contact is formed.
    out : array (optional)
        If `out` is supplied as a pre-allocated array of the correct
        shape then it is filled instead of allocating a new one in
        order to increase performance.

    Returns
    -------
    contacts : ndarray
        boolean array of formed contacts
    """
    if out is not None:
        out[:] = d <= radius
    else:
        out = d <= radius
    return out


class Contacts(AnalysisBase):
    """Calculate contacts based observables.

    The standard methods used in this class calculate the fraction of native
    contacts *Q* from a trajectory.


    .. rubric:: Contact API

    By defining your own method it is possible to calculate other observables
    that only depend on the distances and a possible reference distance. The
    **Contact API** prescribes that this method must be a function with call
    signature ``func(r, r0, **kwargs)`` and must be provided in the keyword
    argument `method`.

    Attributes
    ----------
    results.timeseries : numpy.ndarray
        2D array containing *Q* for all refgroup pairs and analyzed frames

    timeseries : numpy.ndarray
        Alias to the :attr:`results.timeseries` attribute.

        .. deprecated:: 2.0.0
           Will be removed in MDAnalysis 3.0.0. Please use
           :attr:`results.timeseries` instead.


    .. versionchanged:: 1.0.0
       ``save()`` method has been removed. Use ``np.savetxt()`` on
       :attr:`Contacts.results.timeseries` instead.
    .. versionchanged:: 1.0.0
        added ``pbc`` attribute to calculate distances using PBC.
    .. versionchanged:: 2.0.0
       :attr:`timeseries` results are now stored in a
       :class:`MDAnalysis.analysis.base.Results` instance.
    .. versionchanged:: 2.2.0
       :class:`Contacts` accepts both AtomGroup and string for `select`
    .. versionchanged:: 2.9.0
       Introduced :meth:`get_supported_backends` allowing
       for parallel execution on :mod:`multiprocessing`
       and :mod:`dask` backends.
    """

    _analysis_algorithm_is_parallelizable = True

    @classmethod
    def get_supported_backends(cls):
        return (
            "serial",
            "multiprocessing",
            "dask",
        )

    def __init__(
        self,
        u,
        select,
        refgroup,
        method="hard_cut",
        radius=4.5,
        pbc=True,
        kwargs=None,
        **basekwargs,
    ):
        """
        Parameters
        ----------
        u : Universe
            trajectory
        select : tuple(AtomGroup, AtomGroup) | tuple(string, string)
            two contacting groups that change over time
        refgroup : tuple(AtomGroup, AtomGroup)
            two contacting atomgroups in their reference conformation. This
            can also be a list of tuples containing different atom groups
        radius : float, optional (4.5 Angstroms)
            radius within which contacts exist in refgroup
        method : string | callable (optional)
            Can either be one of ``['hard_cut' , 'soft_cut', 'radius_cut']`` or a callable
            with call signature ``func(r, r0, **kwargs)`` (the "Contacts API").
        pbc : bool (optional)
            Uses periodic boundary conditions to calculate distances if set to ``True``; the
            default is ``True``.
        kwargs : dict, optional
            dictionary of additional kwargs passed to `method`. Check
            respective functions for reasonable values.
        verbose : bool (optional)
             Show detailed progress of the calculation if set to ``True``; the
             default is ``False``.

        Attributes
        ----------
        n_initial_contacts : int
             Total number of initial contacts.
        r0 : list[numpy.ndarray]
             List of distance arrays between reference groups.

        Notes
        -----

        .. versionchanged:: 1.0.0
           Changed `selection` keyword to `select`
        """
        self.u = u
        super(Contacts, self).__init__(self.u.trajectory, **basekwargs)

        self.fraction_kwargs = kwargs if kwargs is not None else {}

        if method == "hard_cut":
            self.fraction_contacts = hard_cut_q
        elif method == "soft_cut":
            self.fraction_contacts = soft_cut_q
        elif method == "radius_cut":
            self.fraction_contacts = functools.partial(
                radius_cut_q, radius=radius
            )
        else:
            if not callable(method):
                raise ValueError("method has to be callable")
            self.fraction_contacts = method

        self.select = select

        self.grA, self.grB = (self._get_atomgroup(u, sel) for sel in select)

        self.pbc = pbc

        # contacts formed in reference
        self.r0 = []
        self.initial_contacts = []

        # get dimensions via partial for parallelization compatibility
        self._get_box = functools.partial(self._get_box_func, pbc=self.pbc)

        if isinstance(refgroup[0], AtomGroup):
            refA, refB = refgroup
            self.r0.append(
                distance_array(
                    refA.positions,
                    refB.positions,
                    box=self._get_box(refA.universe),
                )
            )
            self.initial_contacts.append(contact_matrix(self.r0[-1], radius))

        else:
            for refA, refB in refgroup:
                self.r0.append(
                    distance_array(
                        refA.positions,
                        refB.positions,
                        box=self._get_box(refA.universe),
                    )
                )
                self.initial_contacts.append(
                    contact_matrix(self.r0[-1], radius)
                )

        self.n_initial_contacts = self.initial_contacts[0].sum()

    @staticmethod
    def _get_atomgroup(u, sel):
        select_error_message = (
            "selection must be either string or a "
            "static AtomGroup. Updating AtomGroups "
            "are not supported."
        )
        if isinstance(sel, str):
            return u.select_atoms(sel)
        elif isinstance(sel, AtomGroup):
            if isinstance(sel, UpdatingAtomGroup):
                raise TypeError(select_error_message)
            else:
                return sel
        else:
            raise TypeError(select_error_message)

    @staticmethod
    def _get_box_func(ts, pbc):
        """Retrieve the dimensions of the simulation box based on PBC.

        Parameters
        ----------
        ts : Timestep
            The current timestep of the simulation, which contains the
            box dimensions.
        pbc : bool
            A flag indicating whether periodic boundary conditions (PBC)
            are enabled. If `True`, the box dimensions are returned,
            else returns `None`.

        Returns
        -------
        box_dimensions : ndarray or None
            The dimensions of the simulation box as a NumPy array if PBC
            is True, else returns `None`.
        """
        return ts.dimensions if pbc else None

    def _prepare(self):
        self.results.timeseries = np.empty((self.n_frames, len(self.r0) + 1))

    def _single_frame(self):
        self.results.timeseries[self._frame_index][0] = self._ts.frame

        # compute distance array for a frame
        d = distance_array(
            self.grA.positions, self.grB.positions, box=self._get_box(self._ts)
        )

        for i, (initial_contacts, r0) in enumerate(
            zip(self.initial_contacts, self.r0), 1
        ):
            # select only the contacts that were formed in the reference state
            r = d[initial_contacts]
            r0 = r0[initial_contacts]
            q = self.fraction_contacts(r, r0, **self.fraction_kwargs)
            self.results.timeseries[self._frame_index][i] = q

    @property
    def timeseries(self):
        wmsg = (
            "The `timeseries` attribute was deprecated in MDAnalysis "
            "2.0.0 and will be removed in MDAnalysis 3.0.0. Please use "
            "`results.timeseries` instead"
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.timeseries

    def _get_aggregator(self):
        return ResultsGroup(lookup={"timeseries": ResultsGroup.ndarray_vstack})


def _new_selections(u_orig, selections, frame):
    """create stand alone AGs from selections at frame"""
    u = MDAnalysis.Universe(u_orig.filename, u_orig.trajectory.filename)
    u.trajectory[frame]
    return [u.select_atoms(s) for s in selections]


def q1q2(u, select="all", radius=4.5):
    """Perform a q1-q2 analysis.

    Compares native contacts between the starting structure and final structure
    of a trajectory :footcite:p:`Franklin2007`.

    Parameters
    ----------
    u : Universe
        Universe with a trajectory
    select : string, optional
        atoms to do analysis on
    radius : float, optional
        distance at which contact is formed

    Returns
    -------
    contacts : :class:`Contacts`
        Contact Analysis that is set up for a q1-q2 analysis


    .. versionchanged:: 1.0.0
       Changed `selection` keyword to `select`
       Support for setting ``start``, ``stop``, and ``step`` has been removed.
       These should now be directly passed to :meth:`Contacts.run`.
    """
    selection = (select, select)
    first_frame_refs = _new_selections(u, selection, 0)
    last_frame_refs = _new_selections(u, selection, -1)
    return Contacts(
        u,
        selection,
        (first_frame_refs, last_frame_refs),
        radius=radius,
        method="radius_cut",
    )



import pytest

from MDAnalysis.analysis.base import AnalysisBase, AnalysisFromFunction
from MDAnalysisTests.analysis.test_base import (
    FrameAnalysis,
    IncompleteAnalysis,
    OldAPIAnalysis,
)
from MDAnalysis.analysis.rms import RMSD, RMSF
from MDAnalysis.analysis.dihedrals import Dihedral, Ramachandran, Janin
from MDAnalysis.analysis.bat import BAT
from MDAnalysis.analysis.gnm import GNMAnalysis
from MDAnalysis.analysis.dssp.dssp import DSSP
from MDAnalysis.analysis.hydrogenbonds.hbond_analysis import (
    HydrogenBondAnalysis,
)
from MDAnalysis.analysis.nucleicacids import NucPairDist
from MDAnalysis.analysis.contacts import Contacts
from MDAnalysis.analysis.density import DensityAnalysis
from MDAnalysis.analysis.lineardensity import LinearDensity
from MDAnalysis.analysis.polymer import PersistenceLength
from MDAnalysis.analysis.rdf import InterRDF, InterRDF_s
from MDAnalysis.lib.util import is_installed


def params_for_cls(cls, exclude: list[str] = None):
    """
    This part contains fixtures for simultaneous testing
    of all available (=installed & supported) backends
    for analysis subclasses.

    If for some reason you want to limit backends,
    simply pass "exclude: list[str]" to the function
    that parametrizes fixture.

    Parameters
    ----------
    exclude : list[str], optional
        list of backends to exclude from parametrization, by default None

    Returns
    -------
    dict
        dictionary with all tested keyword combinations for the run
    """
    exclude = [] if exclude is None else exclude
    possible_backends = cls.get_supported_backends()
    installed_backends = [
        b for b in possible_backends if is_installed(b) and b not in exclude
    ]

    params = [
        pytest.param(
            {"backend": backend, "n_workers": nproc},
        )
        for backend in installed_backends
        for nproc in (2,)
        if backend != "serial"
    ]
    params.extend([{"backend": "serial"}])
    return params


@pytest.fixture(scope="module", params=params_for_cls(FrameAnalysis))
def client_FrameAnalysis(request):
    return request.param


@pytest.fixture(scope="module", params=params_for_cls(AnalysisBase))
def client_AnalysisBase(request):
    return request.param


@pytest.fixture(scope="module", params=params_for_cls(AnalysisFromFunction))
def client_AnalysisFromFunction(request):
    return request.param


@pytest.fixture(
    scope="module",
    params=params_for_cls(AnalysisFromFunction, exclude=["multiprocessing"]),
)
def client_AnalysisFromFunctionAnalysisClass(request):
    return request.param


@pytest.fixture(scope="module", params=params_for_cls(IncompleteAnalysis))
def client_IncompleteAnalysis(request):
    return request.param


@pytest.fixture(scope="module", params=params_for_cls(OldAPIAnalysis))
def client_OldAPIAnalysis(request):
    return request.param


# MDAnalysis.analysis.rms


@pytest.fixture(scope="module", params=params_for_cls(RMSD))
def client_RMSD(request):
    return request.param


@pytest.fixture(scope="module", params=params_for_cls(RMSF))
def client_RMSF(request):
    return request.param


# MDAnalysis.analysis.dihedrals


@pytest.fixture(scope="module", params=params_for_cls(Dihedral))
def client_Dihedral(request):
    return request.param


@pytest.fixture(scope="module", params=params_for_cls(Ramachandran))
def client_Ramachandran(request):
    return request.param


@pytest.fixture(scope="module", params=params_for_cls(Janin))
def client_Janin(request):
    return request.param


# MDAnalysis.analysis.gnm


@pytest.fixture(scope="module", params=params_for_cls(GNMAnalysis))
def client_GNMAnalysis(request):
    return request.param


# MDAnalysis.analysis.bat


@pytest.fixture(scope="module", params=params_for_cls(BAT))
def client_BAT(request):
    return request.param


# MDAnalysis.analysis.dssp.dssp


@pytest.fixture(scope="module", params=params_for_cls(DSSP))
def client_DSSP(request):
    return request.param


# MDAnalysis.analysis.hydrogenbonds


@pytest.fixture(scope="module", params=params_for_cls(HydrogenBondAnalysis))
def client_HydrogenBondAnalysis(request):
    return request.param


# MDAnalysis.analysis.nucleicacids


@pytest.fixture(scope="module", params=params_for_cls(NucPairDist))
def client_NucPairDist(request):
    return request.param


# MDAnalysis.analysis.contacts


@pytest.fixture(scope="module", params=params_for_cls(Contacts))
def client_Contacts(request):
    return request.param


# MDAnalysis.analysis.density


@pytest.fixture(scope="module", params=params_for_cls(DensityAnalysis))
def client_DensityAnalysis(request):
    return request.param


# MDAnalysis.analysis.lineardensity


@pytest.fixture(scope="module", params=params_for_cls(LinearDensity))
def client_LinearDensity(request):
    return request.param


# MDAnalysis.analysis.polymer


@pytest.fixture(scope="module", params=params_for_cls(PersistenceLength))
def client_PersistenceLength(request):
    return request.param


# MDAnalysis.analysis.rdf


@pytest.fixture(scope="module", params=params_for_cls(InterRDF))
def client_InterRDF(request):
    return request.param


@pytest.fixture(scope="module", params=params_for_cls(InterRDF_s))
def client_InterRDF_s(request):
    return request.param



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
DESRES Molecular Structure file format topology parser
======================================================

Classes to read a topology from a DESRES_ Molecular Structure file
format (DMS_) coordinate files (as used by the Desmond_ MD package).

.. _DESRES: http://www.deshawresearch.com
.. _Desmond: http://www.deshawresearch.com/resources_desmond.html
.. _DMS: http://www.deshawresearch.com/Desmond_Users_Guide-0.7.pdf

Classes
-------

.. autoclass:: DMSParser
   :members:
   :inherited-members:

"""
import numpy as np
import sqlite3
import os

from .base import TopologyReaderBase, change_squash
from ..core.topology import Topology
from ..core.topologyattrs import (
    Atomids,
    Atomnames,
    Bonds,
    Charges,
    ChainIDs,
    Masses,
    Resids,
    Resnums,
    Resnames,
    Segids,
    AtomAttr,  # for custom Attributes
)


class Atomnums(AtomAttr):
    """The number for each Atom"""

    attrname = "atomnums"
    singular = "atomnum"


class DMSParser(TopologyReaderBase):
    """Read a topology from a DESRES_ Molecular Structure file.

    Format (DMS_) coordinate files (as used by the Desmond_ MD package).

    Reads the following attributes:
      Atom:
        - Atomids
        - Atomnums
        - Atomnames
        - Masses
        - Charges
        - Chainids
      Residue:
        - Resnames
        - Resids
      Segment:
        - Segids

    .. note::

        By default, atomtypes will be guessed on Universe creation.
        This may change in release 3.0.
        See :ref:`Guessers` for more information.

    .. _DESRES: http://www.deshawresearch.com
    .. _Desmond: http://www.deshawresearch.com/resources_desmond.html
    .. _DMS: http://www.deshawresearch.com/Desmond_Users_Guide-0.7.pdf
    .. versionchanged:: 2.8.0
        Removed type guessing (attributes guessing takes place now
        through universe.guess_TopologyAttrs() API).

    """

    format = "DMS"

    def parse(self, **kwargs):
        """Parse DMS file *filename* and return the Topology object"""
        # Fix by SB: Needed because sqlite3.connect does not raise anything
        # if file is not there
        if not os.path.isfile(self.filename):
            raise IOError("No such file: {0}".format(self.filename))

        def dict_factory(cursor, row):
            """
            Fetch SQL records as dictionaries, rather than the default tuples.
            """
            d = {}
            for idx, col in enumerate(cursor.description):
                d[col[0]] = row[idx]
            return d

        attrs = {}

        # Row factories for different data types
        facs = {
            np.int32: lambda c, r: r[0],
            np.float32: lambda c, r: r[0],
            object: lambda c, r: str(r[0].strip()),
        }

        with sqlite3.connect(self.filename) as con:
            # Selecting single column, so just strip tuple
            for attrname, dt in [
                ("id", np.int32),
                ("anum", np.int32),
                ("mass", np.float32),
                ("charge", np.float32),
                ("name", object),
                ("resname", object),
                ("resid", np.int32),
                ("chain", object),
                ("segid", object),
            ]:
                try:
                    cur = con.cursor()
                    cur.row_factory = facs[dt]
                    cur.execute("SELECT {} FROM particle" "".format(attrname))
                    vals = cur.fetchall()
                except sqlite3.DatabaseError:
                    errmsg = "Failed reading the atoms from DMS Database"
                    raise IOError(errmsg) from None
                else:
                    attrs[attrname] = np.array(vals, dtype=dt)

            try:
                cur.row_factory = dict_factory
                cur.execute("SELECT * FROM bond")
                bonds = cur.fetchall()
            except sqlite3.DatabaseError:
                errmsg = "Failed reading the bonds from DMS Database"
                raise IOError(errmsg) from None
            else:
                bondlist = []
                bondorder = {}
                for b in bonds:
                    desc = tuple(sorted([b["p0"], b["p1"]]))
                    bondlist.append(desc)
                    bondorder[desc] = b["order"]
                attrs["bond"] = bondlist
                attrs["bondorder"] = bondorder

        topattrs = []
        # Bundle in Atom level objects
        for attr, cls in [
            ("id", Atomids),
            ("anum", Atomnums),
            ("mass", Masses),
            ("charge", Charges),
            ("name", Atomnames),
            ("chain", ChainIDs),
        ]:
            topattrs.append(cls(attrs[attr]))

        # Residues
        atom_residx, (res_resids, res_resnums, res_resnames, res_segids) = (
            change_squash(
                (attrs["resid"], attrs["resname"], attrs["segid"]),
                (
                    attrs["resid"],
                    attrs["resid"].copy(),
                    attrs["resname"],
                    attrs["segid"],
                ),
            )
        )

        n_residues = len(res_resids)
        topattrs.append(Resids(res_resids))
        topattrs.append(Resnums(res_resnums))
        topattrs.append(Resnames(res_resnames))

        if any(res_segids) and not any(val is None for val in res_segids):
            res_segidx, (res_segids,) = change_squash(
                (res_segids,), (res_segids,)
            )

            uniq_seg = np.unique(res_segids)
            idx2seg = {idx: res_segids[idx] for idx in res_segidx}
            res_segids = uniq_seg
            nidx = {segid: nidx for nidx, segid in enumerate(uniq_seg)}

            res_segidx = np.array([nidx[idx2seg[idx]] for idx in res_segidx])

            n_segments = len(res_segids)
            topattrs.append(Segids(res_segids))
        else:
            n_segments = 1
            topattrs.append(Segids(np.array(["SYSTEM"], dtype=object)))
            res_segidx = None

        topattrs.append(Bonds(attrs["bond"]))

        top = Topology(
            len(attrs["id"]),
            n_residues,
            n_segments,
            attrs=topattrs,
            atom_resindex=atom_residx,
            residue_segindex=res_segidx,
        )

        return top



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

r"""Radial Distribution Functions --- :mod:`MDAnalysis.analysis.rdf`
====================================================================

This module contains two classes to calculate radial
`pair distribution functions`_ (`radial distribution functions`_ or "RDF").
The RDF :math:`g_{ab}(r)` between types of particles :math:`a` and :math:`b` is

.. _equation-gab:

.. math::

   g_{ab}(r) = \frac{1}{N_{a}} \frac{1}{N_{b}/V} \sum_{i=1}^{N_a} \sum_{j=1}^{N_b}
               \langle \delta(|\mathbf{r}_i - \mathbf{r}_j| - r) \rangle

which is normalized so that the RDF becomes 1 for large separations in a
homogenous system. The RDF effectively counts the average number of :math:`b`
neighbours in a shell at distance :math:`r` around a :math:`a` particle and
represents it as a density.

The radial cumulative distribution function is

.. math::

   G_{ab}(r) = \int_0^r \!\!dr' 4\pi r'^2 g_{ab}(r')

and the average number of :math:`b` particles within radius :math:`r`

.. _equation-countab:

.. math::

   N_{ab}(r) = \rho G_{ab}(r)

(with the appropriate density :math:`\rho`). The latter function can be used to
compute, for instance, coordination numbers such as the number of neighbors in
the first solvation shell :math:`N(r_1)` where :math:`r_1` is the position of
the first minimum in :math:`g(r)`.

We provide options for calculating the density of particle :math:`b`
in a shell at distance :math:`r` around a :math:`a` particle, which is

.. _equation-nab:

.. math::
   n_{ab}(r) = \rho g_{ab}(r)

.. _`pair distribution functions`:
   https://en.wikipedia.org/wiki/Pair_distribution_function
.. _`radial distribution functions`:
   https://en.wikipedia.org/wiki/Radial_distribution_function

.. Not Implemented yet:
.. - Structure factor?
.. - Coordination number
"""
import warnings
import numpy as np

from ..lib import distances
from .base import AnalysisBase, ResultsGroup


def nested_array_sum(arrs):
    r"""Custom aggregator for nested arrays

    This function takes a nested list or tuple of NumPy arrays, flattens it
    into a single list, and aggregates the elements at alternating indices
    into two separate arrays. The first array accumulates elements at even
    indices, while the second accumulates elements at odd indices.

    Parameters
    ----------
    arrs : list
        List of arrays or nested lists of arrays

    Returns
    -------
    list of ndarray
        A list containing two NumPy arrays:
        - The first array is the sum of all elements at even indices
          in the sum of flattened arrays.
        - The second array is the sum of all elements at odd indices
          in the sum of flattened arrays.
    """

    def flatten(arr):
        if isinstance(arr, (list, tuple)):
            return [item for sublist in arr for item in flatten(sublist)]
        return [arr]

    flat = flatten(arrs)
    aggregated_arr = [np.zeros_like(flat[0]), np.zeros_like(flat[1])]
    for i in range(len(flat) // 2):
        aggregated_arr[0] += flat[2 * i]  # 0, 2, 4, ...
        aggregated_arr[1] += flat[2 * i + 1]  # 1, 3, 5, ...
    return aggregated_arr


class InterRDF(AnalysisBase):
    r"""Radial distribution function

    :class:`InterRDF` is a tool to calculate average radial distribution
    functions between two groups of atoms. Suppose we have two AtomGroups ``A``
    and ``B``. ``A`` contains atom ``A1``, ``A2``, and ``B`` contains ``B1``,
    ``B2``. Given ``A`` and ``B`` to :class:`InterRDF`, the output will be the
    average of RDFs between ``A1`` and ``B1``, ``A1`` and ``B2``, ``A2`` and
    ``B1``, ``A2`` and ``B2``. A typical application is to calculate the RDF of
    solvent with itself or with another solute.

    The :ref:`radial distribution function<equation-gab>` is calculated by
    histogramming distances between all particles in `g1` and `g2` while taking
    periodic boundary conditions into account via the minimum image
    convention.

    The `exclusion_block` keyword may be used to exclude a set of distances
    from the calculations.

    Results are available in the attributes :attr:`results.rdf`
    and :attr:`results.count`.

    Parameters
    ----------
    g1 : AtomGroup
        First AtomGroup
    g2 : AtomGroup
        Second AtomGroup
    nbins : int
        Number of bins in the histogram
    range : tuple or list
        The size of the RDF
    norm : str, {'rdf', 'density', 'none'}
          For 'rdf' calculate :math:`g_{ab}(r)`. For
          'density' the :ref:`single particle density<equation-nab>`
          :math:`n_{ab}(r)` is computed. 'none' computes the number of
          particles occurences in each spherical shell.

          .. versionadded:: 2.3.0

    exclusion_block : tuple
        A tuple representing the tile to exclude from the distance array.
    exclude_same : str
        Will exclude pairs of atoms that share the same "residue", "segment", or "chain".
        Those are the only valid values. This is intended to remove atoms that are
        spatially correlated due to direct bonded connections.
    verbose : bool
        Show detailed progress of the calculation if set to `True`
    backend : {'serial', 'OpenMP', 'distopia'}, optional
        Keyword selecting the type of acceleration of the distance calculations.
        Default is 'serial'.

        .. versionadded:: 2.10.0

    Attributes
    ----------
    results.bins : numpy.ndarray
       :class:`numpy.ndarray` of the centers of the `nbins` histogram
       bins.

       .. versionadded:: 2.0.0

    bins : numpy.ndarray
       Alias to the :attr:`results.bins` attribute.

       .. deprecated:: 2.0.0
           This attribute will be removed in 3.0.0.
           Use :attr:`results.bins` instead.

    results.edges : numpy.ndarray

      :class:`numpy.ndarray` of the `nbins + 1` edges of the histogram
      bins.

       .. versionadded:: 2.0.0

    edges : numpy.ndarray

       Alias to the :attr:`results.edges` attribute.

       .. deprecated:: 2.0.0
           This attribute will be removed in 3.0.0.
           Use :attr:`results.edges` instead.

    results.rdf : numpy.ndarray
      :class:`numpy.ndarray` of the :ref:`radial distribution
      function<equation-gab>` values for the :attr:`results.bins`.

       .. versionadded:: 2.0.0

    rdf : numpy.ndarray
       Alias to the :attr:`results.rdf` attribute.

       .. deprecated:: 2.0.0
           This attribute will be removed in 3.0.0.
           Use :attr:`results.rdf` instead.

    results.count : numpy.ndarray
      :class:`numpy.ndarray` representing the radial histogram, i.e.,
      the raw counts, for all :attr:`results.bins`.

       .. versionadded:: 2.0.0

    count : numpy.ndarray
       Alias to the :attr:`results.count` attribute.

       .. deprecated:: 2.0.0
           This attribute will be removed in 3.0.0.
           Use :attr:`results.count` instead.

    Example
    -------
    First create the :class:`InterRDF` object, by supplying two
    AtomGroups then use the :meth:`run` method ::

      rdf = InterRDF(ag1, ag2)
      rdf.run()

    Results are available through the :attr:`results.bins` and
    :attr:`results.rdf` attributes::

      plt.plot(rdf.results.bins, rdf.results.rdf)

    The `exclusion_block` keyword allows the masking of pairs from
    within the same molecule. For example, if there are 7 of each
    atom in each molecule, the exclusion mask ``(7, 7)`` can be used.


    .. versionadded:: 0.13.0

    .. versionchanged:: 1.0.0
       Support for the `start`, `stop`, and `step` keywords has been
       removed. These should instead be passed to :meth:`InterRDF.run`.

    .. versionchanged:: 2.0.0
       Store results as attributes `bins`, `edges`, `rdf` and `count`
       of the `results` attribute of
       :class:`~MDAnalysis.analysis.AnalysisBase`.

    .. versionchanged:: 2.10.0
       Enabled **parallel execution** with the ``multiprocessing`` and ``dask``
       backends; use the new method :meth:`get_supported_backends` to see all
       supported backends.
    """

    @classmethod
    def get_supported_backends(cls):
        return (
            "serial",
            "multiprocessing",
            "dask",
        )

    _analysis_algorithm_is_parallelizable = True

    def __init__(
        self,
        g1,
        g2,
        nbins=75,
        range=(0.0, 15.0),
        norm="rdf",
        exclusion_block=None,
        exclude_same=None,
        backend="serial",
        **kwargs,
    ):
        super(InterRDF, self).__init__(g1.universe.trajectory, **kwargs)
        self.g1 = g1
        self.g2 = g2
        self.norm = str(norm).lower()

        self.rdf_settings = {"bins": nbins, "range": range}
        self._exclusion_block = exclusion_block
        if exclude_same is not None and exclude_same not in [
            "residue",
            "segment",
            "chain",
        ]:
            raise ValueError(
                "The exclude_same argument to InterRDF must be None, 'residue', 'segment' "
                "or 'chain'."
            )
        if exclude_same is not None and exclusion_block is not None:
            raise ValueError(
                "The exclude_same argument to InterRDF cannot be used with exclusion_block."
            )
        name_to_attr = {
            "residue": "resindices",
            "segment": "segindices",
            "chain": "chainIDs",
        }
        self.exclude_same = name_to_attr.get(exclude_same)

        if self.norm not in ["rdf", "density", "none"]:
            raise ValueError(
                f"'{self.norm}' is an invalid norm. "
                "Use 'rdf', 'density' or 'none'."
            )

        self.backend = backend

    def _prepare(self):
        # Empty histogram to store the RDF
        count, edges = np.histogram([-1], **self.rdf_settings)
        count = count.astype(np.float64)
        count *= 0.0
        self.results.count = count
        self.results.edges = edges
        self.results.bins = 0.5 * (edges[:-1] + edges[1:])

        if self.norm == "rdf":
            # Cumulative volume for rdf normalization
            self.results.volume_cum = 0
        # Set the max range to filter the search radius
        self._maxrange = self.rdf_settings["range"][1]

    def _single_frame(self):
        pairs, dist = distances.capped_distance(
            self.g1.positions,
            self.g2.positions,
            self._maxrange,
            box=self._ts.dimensions,
            backend=self.backend,
        )
        # Maybe exclude same molecule distances
        if self._exclusion_block is not None:
            idxA = pairs[:, 0] // self._exclusion_block[0]
            idxB = pairs[:, 1] // self._exclusion_block[1]
            mask = np.where(idxA != idxB)[0]
            dist = dist[mask]

        if self.exclude_same is not None:
            # Ignore distances between atoms in the same attribute
            attr_ix_a = getattr(self.g1, self.exclude_same)[pairs[:, 0]]
            attr_ix_b = getattr(self.g2, self.exclude_same)[pairs[:, 1]]
            mask = np.where(attr_ix_a != attr_ix_b)[0]
            dist = dist[mask]

        count, _ = np.histogram(dist, **self.rdf_settings)
        self.results.count += count

        if self.norm == "rdf":
            self.results.volume_cum += self._ts.volume

    def _get_aggregator(self):
        return ResultsGroup(
            lookup={
                "count": ResultsGroup.ndarray_sum,
                "volume_cum": ResultsGroup.ndarray_sum,
                "bins": ResultsGroup.ndarray_sum,
                "edges": ResultsGroup.ndarray_mean,
            }
        )

    def _conclude(self):
        norm = self.n_frames
        if self.norm in ["rdf", "density"]:
            # Volume in each radial shell
            vols = np.power(self.results.edges, 3)
            norm *= 4 / 3 * np.pi * np.diff(vols)

        if self.norm == "rdf":
            # Number of each selection
            nA = len(self.g1)
            nB = len(self.g2)
            N = nA * nB

            # If we had exclusions, take these into account
            if self._exclusion_block:
                xA, xB = self._exclusion_block
                nblocks = nA / xA
                N -= xA * xB * nblocks

            # Average number density
            self.volume_cum = self.results.volume_cum
            box_vol = self.volume_cum / self.n_frames
            norm *= N / box_vol

        self.results.rdf = self.results.count / norm

    @property
    def edges(self):
        wmsg = (
            "The `edges` attribute was deprecated in MDAnalysis 2.0.0 "
            "and will be removed in MDAnalysis 3.0.0. Please use "
            "`results.bins` instead"
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.edges

    @property
    def count(self):
        wmsg = (
            "The `count` attribute was deprecated in MDAnalysis 2.0.0 "
            "and will be removed in MDAnalysis 3.0.0. Please use "
            "`results.bins` instead"
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.count

    @property
    def bins(self):
        wmsg = (
            "The `bins` attribute was deprecated in MDAnalysis 2.0.0 "
            "and will be removed in MDAnalysis 3.0.0. Please use "
            "`results.bins` instead"
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.bins

    @property
    def rdf(self):
        wmsg = (
            "The `rdf` attribute was deprecated in MDAnalysis 2.0.0 "
            "and will be removed in MDAnalysis 3.0.0. Please use "
            "`results.rdf` instead"
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.rdf


class InterRDF_s(AnalysisBase):
    r"""Site-specific radial distribution function

    Calculates site-specific radial distribution
    functions. Instead of two groups of atoms it takes as input a list of
    pairs of AtomGroup, ``[[A, B], [C, D], ...]``. Given the same ``A`` and
    ``B`` to
    :class:`InterRDF_s`, the output will be a list of individual RDFs between
    ``A1`` and ``B1``, ``A1`` and ``B2``, ``A2`` and ``B1``, ``A2`` and ``B2``
    (and
    similarly for ``C`` and ``D``). These site-specific radial distribution
    functions are typically calculated if one is interested in the solvation
    shells of a ligand in a binding site or the solvation of specific residues
    in a protein.

    Parameters
    ----------
    u : Universe
        a Universe that contains atoms in `ags`

       .. deprecated:: 2.3.0
           This parameter is superflous and will be removed in
           MDAnalysis 3.0.0.

    ags : list
        a list of pairs of :class:`~MDAnalysis.core.groups.AtomGroup`
        instances
    nbins : int
        Number of bins in the histogram
    range : tuple or list
        The size of the RDF
    norm : str, {'rdf', 'density', 'none'}
        For 'rdf' calculate :math:`g_{ab}(r)`. For
        'density' the :ref:`single particle density<equation-nab>`
        :math:`n_{ab}(r)` is computed. 'none' computes the number of
        particles occurences in each spherical shell.

        .. versionadded:: 2.3.0

    density : bool
        `False`: calculate :math:`g_{ab}(r)`; `True`: calculate
        the true :ref:`single particle density<equation-nab>`
        :math:`n_{ab}(r)`. `density` overwrites the `norm` parameter.

        .. versionadded:: 1.0.1

            This keyword was available since 0.19.0 but was not
            documented. Furthermore, it had the opposite
            meaning. Since 1.0.1 it is officially supported as
            documented.

        .. deprecated:: 2.3.0
            Instead of `density=True` use `norm='density'`

    backend : {'serial', 'OpenMP', 'distopia'}, optional
        Keyword selecting the type of acceleration of the distance calculations.
        Default is 'serial'.

        .. versionadded:: 2.10.0

    Attributes
    ----------
    results.bins : numpy.ndarray
        :class:`numpy.ndarray` of the centers of the `nbins` histogram
        bins; all individual site-specific RDFs have the same bins.

        .. versionadded:: 2.0.0

    bins : numpy.ndarray
        Alias to the :attr:`results.bins` attribute.

        .. deprecated:: 2.0.0
            This attribute will be removed in 3.0.0.
            Use :attr:`results.bins` instead.

    results.edges : numpy.ndarray
        array of the ``nbins + 1`` edges of the histogram
        bins; all individual site-specific RDFs have the same bins.

        .. versionadded:: 2.0.0

    edges : numpy.ndarray
        Alias to the :attr:`results.edges` attribute.

        .. deprecated:: 2.0.0
            This attribute will be removed in 3.0.0.
            Use :attr:`results.edges` instead.

    results.rdf : list
        :class:`list` of the site-specific :ref:`radial distribution
        functions<equation-gab>` if `norm='rdf'` or :ref:`density
        functions<equation-nab>` for the :attr:`bins`
        if `norm='density'`. The list contains
        ``len(ags)`` entries. Each entry for the ``i``-th pair `[A, B]
        = ags[i]` in `ags` is a :class:`numpy.ndarray` with shape
        ``(len(A), len(B))``, i.e., a stack of RDFs. For example,
        ``results.rdf[i][0, 2]`` is the RDF between atoms ``A[0]``
        and ``B[2]``.

        .. versionadded:: 2.0.0

    rdf : list
        Alias to the :attr:`results.rdf` attribute.

        .. deprecated:: 2.0.0
            This attribute will be removed in 3.0.0.
            Use :attr:`results.rdf` instead.

    results.count : list
        :class:`list` of the site-specific radial histograms, i.e., the
        raw counts, for all :attr:`results.bins`. The data have the same
        structure as :attr:`results.rdf` except that the arrays contain
        the raw counts.

        .. versionadded:: 2.0.0

    count : list
        Alias to the :attr:`results.count` attribute.

        .. deprecated:: 2.0.0
            This attribute will be removed in 3.0.0.
            Use :attr:`results.count` instead.

    results.cdf : list
        :class:`list` of the site-specific :ref:`cumulative
        counts<equation-countab>`, for all :attr:`results.bins`. The data
        have the same structure as :attr:`results.rdf` except that the arrays
        contain the cumulative counts.

        This attribute only exists after :meth:`get_cdf` has been run.

        .. versionadded:: 2.0.0

    cdf : list
        Alias to the :attr:`results.cdf` attribute.

        .. deprecated:: 2.0.0
            This attribute will be removed in 3.0.0.
            Use :attr:`results.cdf` instead.

    Example
    -------
    First create the :class:`InterRDF_s` object, by supplying one Universe and
    one list of pairs of AtomGroups, then use the :meth:`~InterRDF_s.run`
    method::

      from MDAnalysisTests.datafiles import GRO_MEMPROT, XTC_MEMPROT
      u = mda.Universe(GRO_MEMPROT, XTC_MEMPROT)

      s1 = u.select_atoms('name ZND and resid 289')
      s2 = u.select_atoms('(name OD1 or name OD2) and resid 51 and sphzone 5.0 (resid 289)')
      s3 = u.select_atoms('name ZND and (resid 291 or resid 292)')
      s4 = u.select_atoms('(name OD1 or name OD2) and sphzone 5.0 (resid 291)')
      ags = [[s1, s2], [s3, s4]]

      rdf = InterRDF_s(u, ags)
      rdf.run()

    Results are available through the :attr:`results.bins`
    and :attr:`results.rdf` attributes::

      plt.plot(rdf.results.bins, rdf.results.rdf[0][0, 0])

    (Which plots the rdf between the first atom in ``s1`` and the first atom in
    ``s2``)

    To generate the *cumulative distribution function* (cdf) in the sense of
    "particles within radius :math:`r`", i.e., :math:`N_{ab}(r)`, use the
    :meth:`~InterRDF_s.get_cdf` method ::

      cdf = rdf.get_cdf()

    Results are available through the :attr:`results.cdf` attribute::

      plt.plot(rdf.results.bins, rdf.results.cdf[0][0, 0])

    (Which plots the cdf between the first atom in ``s1`` and the first atom in
    ``s2``)


    .. versionadded:: 0.19.0

    .. versionchanged:: 1.0.0
       Support for the `start`, `stop`, and `step` keywords has been
       removed. These should instead be passed to :meth:`InterRDF_s.run`.

    .. versionchanged:: 2.0.0
       Store results as attributes `bins`, `edges`, `rdf`, `count`
       and `cdf` of the `results` attribute
       of :class:`~MDAnalysis.analysis.AnalysisBase`.

    .. versionchanged:: 2.3.0
       Introduce `norm` and `exclusion_blocks` attributes.
    .. deprecated:: 2.3.0
       Instead of `density=True` use `norm='density'`
    .. deprecated:: 2.3.0
       The `universe` parameter is superflous.
    .. versionchanged:: 2.10.0
       Enabled **parallel execution** with the ``multiprocessing`` and ``dask``
       backends; use the new method :meth:`get_supported_backends` to see all
       supported backends.
    """

    @classmethod
    def get_supported_backends(cls):
        return (
            "serial",
            "multiprocessing",
            "dask",
        )

    _analysis_algorithm_is_parallelizable = True

    def _get_aggregator(self):
        return ResultsGroup(
            lookup={
                "count": nested_array_sum,
                "volume_cum": ResultsGroup.ndarray_sum,
                "bins": ResultsGroup.ndarray_mean,
                "edges": ResultsGroup.ndarray_mean,
            }
        )

    def __init__(
        self,
        u,
        ags,
        nbins=75,
        range=(0.0, 15.0),
        norm="rdf",
        density=False,
        backend="serial",
        **kwargs,
    ):
        super(InterRDF_s, self).__init__(
            ags[0][0].universe.trajectory, **kwargs
        )

        warnings.warn(
            "The `u` attribute is superflous and will be removed "
            "in MDAnalysis 3.0.0.",
            DeprecationWarning,
        )

        self.ags = ags
        self.norm = str(norm).lower()
        self.rdf_settings = {"bins": nbins, "range": range}

        if self.norm not in ["rdf", "density", "none"]:
            raise ValueError(
                f"'{self.norm}' is an invalid norm. "
                "Use 'rdf', 'density' or 'none'."
            )

        if density:
            warnings.warn(
                "The `density` attribute was deprecated in "
                "MDAnalysis 2.3.0 and will be removed in "
                "MDAnalysis 3.0.0. Please use `norm=density` "
                "instead.",
                DeprecationWarning,
            )
            self.norm = "density"

        self.backend = backend

    def _prepare(self):
        count, edges = np.histogram([-1], **self.rdf_settings)
        self.results.count = [
            np.zeros((ag1.n_atoms, ag2.n_atoms, len(count)), dtype=np.float64)
            for ag1, ag2 in self.ags
        ]
        self.results.edges = edges
        self.results.bins = 0.5 * (edges[:-1] + edges[1:])

        if self.norm == "rdf":
            # Cumulative volume for rdf normalization
            self.results.volume_cum = 0
        self._maxrange = self.rdf_settings["range"][1]

    def _single_frame(self):
        for i, (ag1, ag2) in enumerate(self.ags):
            pairs, dist = distances.capped_distance(
                ag1.positions,
                ag2.positions,
                self._maxrange,
                box=self._ts.dimensions,
                backend=self.backend,
            )

            # Fast manual implementation for distance histogram (equidistant bins)
            nbins = self.rdf_settings["bins"]
            rmin, rmax = self.rdf_settings["range"]
            bin_indices = (dist - rmin) * nbins / (rmax - rmin)
            bin_indices = bin_indices.astype(np.int64)
            counts = (bin_indices >= 0) & (bin_indices < nbins)
            counts = counts.astype(np.int64)
            idx1s = pairs[:, 0]
            idx2s = pairs[:, 1]
            self.results.count[i][idx1s, idx2s, bin_indices] += counts

        if self.norm == "rdf":
            self.results.volume_cum += self._ts.volume

    def _conclude(self):
        norm = self.n_frames
        if self.norm in ["rdf", "density"]:
            # Volume in each radial shell
            vols = np.power(self.results.edges, 3)
            norm *= 4 / 3 * np.pi * np.diff(vols)

        if self.norm == "rdf":
            # Average number density
            self.volume_cum = self.results.volume_cum
            norm *= 1 / (self.volume_cum / self.n_frames)

        # Empty lists to restore indices, RDF
        self.results.indices = []
        self.results.rdf = []

        for i, (ag1, ag2) in enumerate(self.ags):
            # Number of each selection
            self.results.indices.append([ag1.indices, ag2.indices])
            self.results.rdf.append(self.results.count[i] / norm)

    def get_cdf(self):
        r"""Calculate the cumulative counts for all sites.

        This is the :ref:`cumulative count<equation-countab>` within a given
        radius, i.e., :math:`N_{ab}(r)`.

        The result is returned and also stored in the attribute
        :attr:`results.cdf`.

        Returns
        -------
        cdf : list
              list of arrays with the same structure as :attr:`results.rdf`
        """
        self.results.cdf = []

        for count in self.results.count:
            self.results.cdf.append(np.cumsum(count, axis=2) / self.n_frames)

        return self.results.cdf

    @property
    def edges(self):
        wmsg = (
            "The `edges` attribute was deprecated in MDAnalysis 2.0.0 "
            "and will be removed in MDAnalysis 3.0.0. Please use "
            "`results.bins` instead"
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.edges

    @property
    def count(self):
        wmsg = (
            "The `count` attribute was deprecated in MDAnalysis 2.0.0 "
            "and will be removed in MDAnalysis 3.0.0. Please use "
            "`results.bins` instead"
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.count

    @property
    def bins(self):
        wmsg = (
            "The `bins` attribute was deprecated in MDAnalysis 2.0.0 "
            "and will be removed in MDAnalysis 3.0.0. Please use "
            "`results.bins` instead"
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.bins

    @property
    def rdf(self):
        wmsg = (
            "The `rdf` attribute was deprecated in MDAnalysis 2.0.0 "
            "and will be removed in MDAnalysis 3.0.0. Please use "
            "`results.rdf` instead"
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.rdf

    @property
    def cdf(self):
        wmsg = (
            "The `cdf` attribute was deprecated in MDAnalysis 2.0.0 "
            "and will be removed in MDAnalysis 3.0.0. Please use "
            "`results.cdf` instead"
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.cdf



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""Streamplots (2D)  --- :mod:`MDAnalysis.visualization.streamlines`
=================================================================

:Authors: Tyler Reddy and Matthieu Chavent
:Year: 2014
:Copyright: Lesser GNU Public License v2.1+


The :func:`generate_streamlines` function can generate a 2D flow field from a
MD trajectory, for instance, lipid molecules in a flat membrane. It can make
use of multiple cores to perform the analyis in parallel (using
:mod:`multiprocessing`).

See Also
--------
MDAnalysis.visualization.streamlines_3D : streamplots in 3D


.. autofunction:: generate_streamlines

"""
import multiprocessing

import numpy as np
import scipy

try:
    import matplotlib
    import matplotlib.path
except ImportError:
    raise ImportError(
        "2d streamplot module requires: matplotlib.path for its "
        "path.Path.contains_points method. The installation "
        "instructions for the matplotlib module can be found here: "
        "http://matplotlib.org/faq/installing_faq.html?highlight=install"
    ) from None

import MDAnalysis


def produce_grid(tuple_of_limits, grid_spacing):
    """Produce a 2D grid for the simulation system.

    The grid is based on the tuple of Cartesian Coordinate limits calculated in
    an earlier step.

    Parameters
    ----------
    tuple_of_limits : tuple
        ``x_min, x_max, y_min, y_max``
    grid_spacing : float
        grid size in all directions in ångström

    Returns
    -------
    grid : array
       ``numpy.mgrid[x_min:x_max:grid_spacing, y_min:y_max:grid_spacing]``
    """
    x_min, x_max, y_min, y_max = tuple_of_limits
    grid = np.mgrid[x_min:x_max:grid_spacing, y_min:y_max:grid_spacing]
    return grid


def split_grid(grid, num_cores):
    """Split the grid into blocks of vertices.

    Take the overall `grid` for the system and split it into lists of
    square vertices that can be distributed to each core.

    Parameters
    ----------
    grid : numpy.array
        2D array
    num_cores : int
        number of partitions to generate

    Returns
    -------
    list_square_vertex_arrays_per_core : array of arrays
         split the list of square vertices
         ``[[v1,v2,v3,v4],[v1,v2,v3,v4],...,...]`` into roughly equally-sized
         sublists to be distributed over the available cores on the system
    list_parent_index_values : array of arrays
         arrays of `[[row, column], [row, column], ...]`` for each core
    current_row : int
         last row + 1
    current_column : int
         last column + 1

    Note
    ----
    Limited to 2D for now.

    """

    # produce an array containing the cartesian coordinates of all vertices in the grid:
    x_array, y_array = grid
    grid_vertex_cartesian_array = np.dstack((x_array, y_array))
    # the grid_vertex_cartesian_array has N_rows, with each row corresponding to a column of coordinates in the grid (
    # so a given row has shape N_rows, 2); overall shape (N_columns_in_grid, N_rows_in_a_column, 2)
    # although I'll eventually want a pure numpy/scipy/vector-based solution, for now I'll allow loops to simplify the
    #  division of the cartesian coordinates into a list of the squares in the grid
    list_all_squares_in_grid = (
        []
    )  # should eventually be a nested list of all the square vertices in the grid/system
    list_parent_index_values = (
        []
    )  # want an ordered list of assignment indices for reconstructing the grid positions
    # in the parent process
    current_column = 0
    while current_column < grid_vertex_cartesian_array.shape[0] - 1:
        # go through all the columns except the last one and account for the square vertices (the last column
        #  has no 'right neighbour')
        current_row = 0
        while current_row < grid_vertex_cartesian_array.shape[1] - 1:
            # all rows except the top row, which doesn't have a row above it for forming squares
            bottom_left_vertex_current_square = grid_vertex_cartesian_array[
                current_column, current_row
            ]
            bottom_right_vertex_current_square = grid_vertex_cartesian_array[
                current_column + 1, current_row
            ]
            top_right_vertex_current_square = grid_vertex_cartesian_array[
                current_column + 1, current_row + 1
            ]
            top_left_vertex_current_square = grid_vertex_cartesian_array[
                current_column, current_row + 1
            ]
            # append the vertices of this square to the overall list of square vertices:
            list_all_squares_in_grid.append(
                [
                    bottom_left_vertex_current_square,
                    bottom_right_vertex_current_square,
                    top_right_vertex_current_square,
                    top_left_vertex_current_square,
                ]
            )
            list_parent_index_values.append([current_row, current_column])
            current_row += 1
        current_column += 1
    # split the list of square vertices [[v1,v2,v3,v4],[v1,v2,v3,v4],...,...] into roughly equally-sized sublists to
    # be distributed over the available cores on the system:
    list_square_vertex_arrays_per_core = np.array_split(
        list_all_squares_in_grid, num_cores
    )
    list_parent_index_values = np.array_split(
        list_parent_index_values, num_cores
    )
    return [
        list_square_vertex_arrays_per_core,
        list_parent_index_values,
        current_row,
        current_column,
    ]


# some private utility functions for trajectory iteration
def _produce_list_indices_point_in_polygon_this_frame(
    vertex_coord_list, relevant_particle_coordinate_array_xy
):
    """
    Produce a list of indices for particles of interest in the current frame.
    The list is ordered according to the order of the squares in the grid.
    Each entry in the list is a tuple of indices for the particles in that square.
    The list is the same length as the number of squares in the grid.
    """
    list_indices_point_in_polygon = []
    for square_vertices in vertex_coord_list:
        path_object = matplotlib.path.Path(square_vertices)
        index_list_in_polygon = np.where(
            path_object.contains_points(relevant_particle_coordinate_array_xy)
        )
        list_indices_point_in_polygon.append(index_list_in_polygon)
    return list_indices_point_in_polygon


def _produce_list_centroids_this_frame(
    list_indices_in_polygon, relevant_particle_coordinate_array_xy
):
    """
    Produce a list of centroids for the particles in the current frame.
    The list is ordered according to the order of the squares in the grid.
    Each entry in the list is a numpy array of the centroid coordinates for the particles in that square.
    The list is the same length as the number of squares in the grid.
    If there are no particles in a square, the entry is None.
    """
    list_centroids_this_frame = []
    for indices in list_indices_in_polygon:
        if (
            not indices[0].size > 0
        ):  # if there are no particles of interest in this particular square
            list_centroids_this_frame.append(None)
        else:
            current_coordinate_array_in_square = (
                relevant_particle_coordinate_array_xy[indices]
            )
            current_square_indices_centroid = np.average(
                current_coordinate_array_in_square, axis=0
            )
            list_centroids_this_frame.append(current_square_indices_centroid)
    return list_centroids_this_frame  # a list of numpy xy centroid arrays for this frame


def per_core_work(
    topology_file_path,
    trajectory_file_path,
    list_square_vertex_arrays_this_core,
    MDA_selection,
    start_frame,
    end_frame,
    reconstruction_index_list,
    maximum_delta_magnitude,
):
    """Run the analysis on one core.

    The code to perform on a given core given the list of square vertices assigned to it.
    """
    # obtain the relevant coordinates for particles of interest
    universe_object = MDAnalysis.Universe(
        topology_file_path, trajectory_file_path
    )
    list_previous_frame_centroids = []
    list_previous_frame_indices = []
    for ts in universe_object.trajectory:
        if ts.frame < start_frame:  # don't start until first specified frame
            continue
        relevant_particle_coordinate_array_xy = universe_object.select_atoms(
            MDA_selection
        ).positions[..., :-1]
        # only 2D / xy coords for now
        # I will need a list of indices for relevant particles falling within each square in THIS frame:
        list_indices_in_squares_this_frame = (
            _produce_list_indices_point_in_polygon_this_frame(
                list_square_vertex_arrays_this_core,
                relevant_particle_coordinate_array_xy,
            )
        )
        # likewise, I will need a list of centroids of particles in each square (same order as above list):
        list_centroids_in_squares_this_frame = (
            _produce_list_centroids_this_frame(
                list_indices_in_squares_this_frame,
                relevant_particle_coordinate_array_xy,
            )
        )
        if (
            list_previous_frame_indices
        ):  # if the previous frame had indices in at least one square I will need to use
            #  those indices to generate the updates to the corresponding centroids in this frame:
            list_centroids_this_frame_using_indices_from_last_frame = (
                _produce_list_centroids_this_frame(
                    list_previous_frame_indices,
                    relevant_particle_coordinate_array_xy,
                )
            )
            # I need to write a velocity of zero if there are any 'empty' squares in either frame:
            xy_deltas_to_write = []
            for square_1_centroid, square_2_centroid in zip(
                list_centroids_this_frame_using_indices_from_last_frame,
                list_previous_frame_centroids,
            ):
                if square_1_centroid is None or square_2_centroid is None:
                    xy_deltas_to_write.append([0, 0])
                else:
                    xy_deltas_to_write.append(
                        np.subtract(
                            square_1_centroid, square_2_centroid
                        ).tolist()
                    )

            # xy_deltas_to_write = np.subtract(np.array(
            # list_centroids_this_frame_using_indices_from_last_frame),np.array(list_previous_frame_centroids))
            xy_deltas_to_write = np.array(xy_deltas_to_write)
            # now filter the array to only contain distances in the range [-8,8] as a placeholder for dealing with PBC
            #  issues (Matthieu seemed to use a limit of 8 as well);
            xy_deltas_to_write = np.clip(
                xy_deltas_to_write,
                -maximum_delta_magnitude,
                maximum_delta_magnitude,
            )

            # with the xy and dx,dy values calculated I need to set the values from this frame to previous frame
            # values in anticipation of the next frame:
            list_previous_frame_centroids = (
                list_centroids_in_squares_this_frame[:]
            )
            list_previous_frame_indices = list_indices_in_squares_this_frame[:]
        else:  # either no points in squares or after the first frame I'll just reset the 'previous' values so they
            # can be used when consecutive frames have proper values
            list_previous_frame_centroids = (
                list_centroids_in_squares_this_frame[:]
            )
            list_previous_frame_indices = list_indices_in_squares_this_frame[:]
        if ts.frame > end_frame:
            break  # stop here
    return list(zip(reconstruction_index_list, xy_deltas_to_write.tolist()))


def generate_streamlines(
    topology_file_path,
    trajectory_file_path,
    grid_spacing,
    MDA_selection,
    start_frame,
    end_frame,
    xmin,
    xmax,
    ymin,
    ymax,
    maximum_delta_magnitude,
    num_cores="maximum",
):
    r"""Produce the x and y components of a 2D streamplot data set.

    Parameters
    ----------
    topology_file_path : str
            Absolute path to the topology file
    trajectory_file_path : str
            Absolute path to the trajectory file. It will normally be desirable
            to filter the trajectory with a tool such as GROMACS
            :program:`g_filter` (see :footcite:p:`Chavent2014`)
    grid_spacing : float
            The spacing between grid lines (angstroms)
    MDA_selection : str
            MDAnalysis selection string
    start_frame : int
            First frame number to parse
    end_frame : int
            Last frame number to parse
    xmin : float
            Minimum coordinate boundary for x-axis (angstroms)
    xmax : float
            Maximum coordinate boundary for x-axis (angstroms)
    ymin : float
            Minimum coordinate boundary for y-axis (angstroms)
    ymax : float
            Maximum coordinate boundary for y-axis (angstroms)
    maximum_delta_magnitude : float
            Absolute value of the largest displacement tolerated for the
            centroid of a group of particles ( angstroms). Values above this
            displacement will not count in the streamplot (treated as
            excessively large displacements crossing the periodic boundary)
    num_cores : int or 'maximum' (optional)
            The number of cores to use. (Default 'maximum' uses all available
            cores)

    Returns
    -------
    dx_array : array of floats
            An array object containing the displacements in the x direction
    dy_array : array of floats
            An array object containing the displacements in the y direction
    average_displacement : float
            :math:`\frac{\sum\sqrt[]{dx^2 + dy^2}}{N}`
    standard_deviation_of_displacement : float
            standard deviation of :math:`\sqrt[]{dx^2 + dy^2}`

    Examples
    --------
    Generate 2D streamlines and plot::

        import matplotlib, matplotlib.pyplot, np
        import MDAnalysis, MDAnalysis.visualization.streamlines

        u1, v1, average_displacement, standard_deviation_of_displacement =
            MDAnalysis.visualization.streamlines.generate_streamlines('testing.gro', 'testing_filtered.xtc',
                    grid_spacing=20, MDA_selection='name PO4', start_frame=2, end_frame=3,
                    xmin=-8.73000049591, xmax= 1225.96008301,
                    ymin= -12.5799999237, ymax=1224.34008789,
                    maximum_delta_magnitude=1.0, num_cores=16)
        x = np.linspace(0, 1200, 61)
        y = np.linspace(0, 1200, 61)
        speed = np.sqrt(u1*u1 + v1*v1)
        fig = matplotlib.pyplot.figure()
        ax = fig.add_subplot(111, aspect='equal')
        ax.set_xlabel('x ($\AA$)')
        ax.set_ylabel('y ($\AA$)')
        ax.streamplot(x, y, u1, v1, density=(10,10), color=speed, linewidth=3*speed/speed.max())
        fig.savefig('testing_streamline.png',dpi=300)

    .. image:: testing_streamline.png


    References
    ----------
    .. footbibliography::


    See Also
    --------
    MDAnalysis.visualization.streamlines_3D.generate_streamlines_3d

    """
    # work out the number of cores to use:
    if num_cores == "maximum":
        num_cores = multiprocessing.cpu_count()  # use all available cores
    else:
        num_cores = num_cores  # use the value specified by the user
        # assert isinstance(num_cores,(int,long)), "The number of specified cores must (of course) be an integer."
    np.seterr(all="warn", over="raise")
    parent_list_deltas = []  # collect all data from child processes here

    def log_result_to_parent(delta_array):
        parent_list_deltas.extend(delta_array)

    tuple_of_limits = (xmin, xmax, ymin, ymax)
    grid = produce_grid(
        tuple_of_limits=tuple_of_limits, grid_spacing=grid_spacing
    )
    (
        list_square_vertex_arrays_per_core,
        list_parent_index_values,
        total_rows,
        total_columns,
    ) = split_grid(grid=grid, num_cores=num_cores)
    pool = multiprocessing.Pool(num_cores)
    for vertex_sublist, index_sublist in zip(
        list_square_vertex_arrays_per_core, list_parent_index_values
    ):
        pool.apply_async(
            per_core_work,
            args=(
                topology_file_path,
                trajectory_file_path,
                vertex_sublist,
                MDA_selection,
                start_frame,
                end_frame,
                index_sublist,
                maximum_delta_magnitude,
            ),
            callback=log_result_to_parent,
        )
    pool.close()
    pool.join()
    dx_array = np.zeros((total_rows, total_columns))
    dy_array = np.zeros((total_rows, total_columns))
    # the parent_list_deltas is shaped like this: [ ([row_index,column_index],[dx,dy]), ... (...),...,]
    for (
        index_array,
        delta_array,
    ) in (
        parent_list_deltas
    ):  # go through the list in the parent process and assign to the
        #  appropriate positions in the dx and dy matrices:
        # build in a filter to replace all values at the cap (currently between -8,8) with 0 to match Matthieu's code
        # (I think eventually we'll reduce the cap to a narrower boundary though)
        index_1 = index_array.tolist()[0]
        index_2 = index_array.tolist()[1]
        if abs(delta_array[0]) == maximum_delta_magnitude:
            dx_array[index_1, index_2] = 0
        else:
            dx_array[index_1, index_2] = delta_array[0]
        if abs(delta_array[1]) == maximum_delta_magnitude:
            dy_array[index_1, index_2] = 0
        else:
            dy_array[index_1, index_2] = delta_array[1]

    # at Matthieu's request, we now want to calculate the average and standard deviation of the displacement values:
    displacement_array = np.sqrt(dx_array**2 + dy_array**2)
    average_displacement = np.average(displacement_array)
    standard_deviation_of_displacement = np.std(displacement_array)

    return (
        dx_array,
        dy_array,
        average_displacement,
        standard_deviation_of_displacement,
    )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""\
Hydrogen bond autocorrelation --- :mod:`MDAnalysis.analysis.hydrogenbonds.hbond_autocorrel`
===========================================================================================

:Author: Richard J. Gowers
:Year: 2014
:Copyright: Lesser GNU Public License v2.1+

.. versionadded:: 0.9.0

.. versionchanged:: 2.0.0

   Module moved from :mod:`MDAnalysis.analysis.hbonds.hbond_autocorrel` to
   :mod:`MDAnalysis.analysis.hydrogenbonds.hbond_autocorrel`.


Description
-----------

Calculates the time autocorrelation function, :math:`C_x(t)`, for the hydrogen
bonds in the selections passed to it.  The population of hydrogen bonds at a
given startpoint, :math:`t_0`, is evaluated based on geometric criteria and
then the lifetime of these bonds is monitored over time.  Multiple passes
through the trajectory are used to build an average of the behaviour.

.. math::
   C_x(t) = \\left \\langle \\frac{h_{ij}(t_0) h_{ij}(t_0 + t)}{h_{ij}(t_0)^2} \\right\\rangle

The subscript :math:`x` refers to the definition of lifetime being used, either
continuous or intermittent.  The continuous definition measures the time that
a particular hydrogen bond remains continuously attached, whilst the
intermittent definition allows a bond to break and then subsequently reform and
be counted again.  The relevent lifetime, :math:`\\tau_x`, can then be found
via integration of this function

.. math::
   \\tau_x = \\int_0^\\infty C_x(t) dt`

For this, the observed behaviour is fitted to a multi exponential function,
using 2 exponents for the continuous lifetime and 3 for the intermittent
lifetime.

.. math::
    C_x(t) = A_1 \\exp( - t / \\tau_1)
    + A_2 \\exp( - t / \\tau_2)
    [+ A_3 \\exp( - t / \\tau_3)]

Where the final pre expoential factor :math:`A_n` is subject to the condition:

.. math::
    A_n = 1 - \\sum\\limits_{i=1}^{n-1} A_i

For further details see :footcite:p:`Gowers2015`.

Input
-----

Three AtomGroup selections representing the **hydrogens**, **donors** and
**acceptors** that you wish to analyse.  Note that the **hydrogens** and
**donors** selections must be aligned, that is **hydrogens[0]** and
**donors[0]** must represent a bonded pair.  For systems such as water,
this will mean that each oxygen appears twice in the **donors** AtomGroup.
The function :func:`find_hydrogen_donors` can be used to construct the donor
AtomGroup
::

  import MDAnalysis as mda
  from MDAnalysis.analysis import hydrogenbonds
  from MDAnalysis.tests.datafiles import waterPSF, waterDCD
  u = mda.Universe(waterPSF, waterDCD)
  hydrogens = u.select_atoms('name H*')
  donors = hydrogenbonds.find_hydrogen_donors(hydrogens)


Note that this requires the Universe to have bond information.  If this isn't
present in the topology file, the
:meth:`MDAnalysis.core.groups.AtomGroup.guess_bonds` method can be used
as so
::

  import MDAnalysis as mda
  from MDAnalysis.analysis import hydrogenbonds
  from MDAnalysis.tests.datafiles import GRO
  # we could load the Universe with guess_bonds=True
  # but this would guess **all** bonds
  u = mda.Universe(GRO)
  water = u.select_atoms('resname SOL and not type DUMMY')
  # guess bonds only within our water atoms
  # this adds the bond information directly to the Universe
  water.guess_bonds()
  hydrogens = water.select_atoms('type H')
  # this is now possible as we guessed the bonds
  donors = hydrogenbonds.find_hydrogen_donors(hydrogens)


The keyword **exclusions** allows a tuple of array addresses to be provided,
(Hidx, Aidx),these pairs of hydrogen-acceptor are then not permitted to be
counted as part of the analysis.  This could be used to exclude the
consideration of hydrogen bonds within the same functional group, or to perform
analysis on strictly intermolecular hydrogen bonding.

Hydrogen bonds are defined on the basis of geometric criteria; a
Hydrogen-Acceptor distance of less then **dist_crit** and a
Donor-Hydrogen-Acceptor angle of greater than **angle_crit**.

The length of trajectory to analyse in ps, **sample_time**, is used to choose
what length to analyse.

Multiple passes, controlled by the keyword **nruns**, through the trajectory
are performed and an average calculated.  For each pass, **nsamples** number
of points along the run are calculated.


Output
------

All results of the analysis are available through the *solution* attribute.
This is a dictionary with the following keys

- *results*:  The raw results of the time autocorrelation function.
- *time*:     Time axis, in ps, for the results.
- *fit*:      Results of the exponential curve fitting procedure. For the
             *continuous* lifetime these are (A1, tau1, tau2), for the
             *intermittent* lifetime these are (A1, A2, tau1, tau2, tau3).
- *tau*:      Calculated time constant from the fit.
- *estimate*: Estimated values generated by the calculated fit.

The *results* and *time* values are only filled after the :meth:`run` method,
*fit*, *tau* and *estimate* are filled after the :meth:`solve` method has been
used.


Worked Example for Polyamide
----------------------------

This example finds the continuous hydrogen bond lifetime between N-H..O in a
polyamide system.  This will use the default geometric definition for hydrogen
bonds of length 3.0 Å and angle of 130 degrees.
It will observe a window of 2.0 ps (`sample_time`) and try to gather 1000
sample point within this time window (this relies upon the trajectory being
sampled frequently enough).  This process is repeated for 20 different start
points to build a better average.

::

  import MDAnalysis as mda
  from MDAnalysis.analysis import hydrogenbonds
  from MDAnalysis.tests.datafiles import TRZ_psf, TRZ
  import matplotlib.pyplot as plt
  # load system
  u = mda.Universe(TRZ_psf, TRZ)
  # select atoms of interest into AtomGroups
  H = u.select_atoms('name Hn')
  N = u.select_atoms('name N')
  O = u.select_atoms('name O')
  # create analysis object
  hb_ac = hydrogenbonds.HydrogenBondAutoCorrel(u,
              acceptors=O, hydrogens=H, donors=N,
              bond_type='continuous',
              sample_time=2.0, nsamples=1000, nruns=20)
  # call run to gather results
  hb_ac.run()
  # attempt to fit results to exponential equation
  hb_ac.solve()
  # grab results from inside object
  tau = hb_ac.solution['tau']
  time = hb_ac.solution['time']
  results = hb_ac.solution['results']
  estimate = hb_ac.solution['estimate']
  # plot to check!
  plt.plot(time, results, 'ro')
  plt.plot(time, estimate)
  plt.show()


Functions and Classes
---------------------

.. autofunction:: find_hydrogen_donors

.. autoclass:: HydrogenBondAutoCorrel
   :members:

"""
import numpy as np
import scipy.optimize
import warnings

from MDAnalysis.lib._cutil import _in2d
from MDAnalysis.lib.log import ProgressBar
from MDAnalysis.lib.distances import capped_distance, calc_angles, calc_bonds
from MDAnalysis.core.groups import requires

from MDAnalysis.due import due, Doi

due.cite(
    Doi("10.1063/1.4922445"),
    description="Hydrogen bonding autocorrelation time",
    path="MDAnalysis.analysis.hydrogenbonds.hbond_autocorrel",
)
del Doi


@requires("bonds")
def find_hydrogen_donors(hydrogens):
    """Returns the donor atom for each hydrogen

    Parameters
    ----------
    hydrogens : AtomGroup
      the hydrogens that will form hydrogen bonds

    Returns
    -------
    donors : AtomGroup
      the donor atom for each hydrogen, found via bond information


    .. versionadded:: 0.20.0
    """
    return sum(h.bonded_atoms[0] for h in hydrogens)


class HydrogenBondAutoCorrel(object):
    """Perform a time autocorrelation of the hydrogen bonds in the system.

    Parameters
    ----------
    universe : Universe
        MDAnalysis Universe that all selections belong to
    hydrogens : AtomGroup
        AtomGroup of Hydrogens which can form hydrogen bonds
    acceptors : AtomGroup
        AtomGroup of all Acceptor atoms
    donors : AtomGroup
        The atoms which are connected to the hydrogens.  This group
        must be identical in length to the hydrogen group and matched,
        ie hydrogens[0] is bonded to donors[0].
        For water, this will mean a donor appears twice in this
        group, once for each hydrogen.
    bond_type : str
        Which definition of hydrogen bond lifetime to consider, either
        'continuous' or 'intermittent'.
    exclusions : ndarray, optional
        Indices of Hydrogen-Acceptor pairs to be excluded.
        With nH and nA Hydrogens and Acceptors, a (nH x nA) array of distances
        is calculated, *exclusions* is used as a mask on this array to exclude
        some pairs.
    angle_crit : float, optional
        The angle (in degrees) which all bonds must be greater than [130.0]
    dist_crit : float, optional
        The maximum distance (in Angstroms) for a hydrogen bond [3.0]
    sample_time : float, optional
        The amount of time, in ps, that you wish to observe hydrogen
        bonds for [100]
    nruns : int, optional
        The number of different start points within the trajectory
        to use [1]
    nsamples : int, optional
        Within each run, the number of frames to analyse [50]
    pbc : bool, optional
        Whether to consider periodic boundaries in calculations [``True``]

    ..versionchanged: 1.0.0
      ``save_results()`` method was removed. You can instead use ``np.savez()``
      on :attr:`HydrogenBondAutoCorrel.solution['time']` and
      :attr:`HydrogenBondAutoCorrel.solution['results']` instead.
    """

    def __init__(
        self,
        universe,
        hydrogens=None,
        acceptors=None,
        donors=None,
        bond_type=None,
        exclusions=None,
        angle_crit=130.0,
        dist_crit=3.0,  # geometric criteria
        sample_time=100,  # expected length of the decay in ps
        time_cut=None,  # cutoff time for intermittent hbonds
        nruns=1,  # number of times to iterate through the trajectory
        nsamples=50,  # number of different points to sample in a run
        pbc=True,
    ):

        # warnings.warn("This class is deprecated, use analysis.hbonds.HydrogenBondAnalysis "
        #              "which has .autocorrelation function",
        #              category=DeprecationWarning)

        self.u = universe
        # check that slicing is possible
        try:
            self.u.trajectory[0]
        except Exception:
            raise ValueError("Trajectory must support slicing") from None

        self.h = hydrogens
        self.a = acceptors
        self.d = donors
        if not len(self.h) == len(self.d):
            raise ValueError(
                "Donors and Hydrogen groups must be identical "
                "length.  Try using `find_hydrogen_donors`."
            )

        if exclusions is not None:
            if len(exclusions[0]) != len(exclusions[1]):
                raise ValueError(
                    "'exclusion' must be two arrays of identical length"
                )
            self.exclusions = np.column_stack(
                (exclusions[0], exclusions[1])
            ).astype(np.intp)
        else:
            self.exclusions = None

        self.bond_type = bond_type
        if self.bond_type not in ["continuous", "intermittent"]:
            raise ValueError(
                "bond_type must be either 'continuous' or 'intermittent'"
            )

        self.a_crit = np.deg2rad(angle_crit)
        self.d_crit = dist_crit
        self.pbc = pbc
        self.sample_time = sample_time
        self.nruns = nruns
        self.nsamples = nsamples
        self._slice_traj(sample_time)
        self.time_cut = time_cut

        self.solution = {
            "results": None,  # Raw results
            "time": None,  # Time axis of raw results
            "fit": None,  # coefficients for fit
            "tau": None,  # integral of exponential fit
            "estimate": None,  # y values of fit against time
        }

    def _slice_traj(self, sample_time):
        """Set up start and end points in the trajectory for the
        different passes
        """
        dt = self.u.trajectory.dt  # frame step size in time
        req_frames = int(sample_time / dt)  # the number of frames required

        n_frames = len(self.u.trajectory)
        if req_frames > n_frames:
            warnings.warn(
                "Number of required frames ({}) greater than the"
                " number of frames in trajectory ({})".format(
                    req_frames, n_frames
                ),
                RuntimeWarning,
            )

        numruns = self.nruns
        if numruns > n_frames:
            numruns = n_frames
            warnings.warn(
                "Number of runs ({}) greater than the number of"
                " frames in trajectory ({})".format(self.nruns, n_frames),
                RuntimeWarning,
            )

        self._starts = np.arange(0, n_frames, n_frames / numruns, dtype=int)
        # limit stop points using clip
        self._stops = np.clip(self._starts + req_frames, 0, n_frames)

        self._skip = req_frames // self.nsamples
        if self._skip == 0:  # If nsamples > req_frames
            warnings.warn(
                "Desired number of sample points too high, using {0}".format(
                    req_frames
                ),
                RuntimeWarning,
            )
            self._skip = 1

    def run(self, force=False):
        """Run all the required passes

        Parameters
        ----------
        force : bool, optional
            Will overwrite previous results if they exist
        """
        # if results exist, don't waste any time
        if self.solution["results"] is not None and not force:
            return

        main_results = np.zeros_like(
            np.arange(self._starts[0], self._stops[0], self._skip),
            dtype=np.float32,
        )
        # for normalising later
        counter = np.zeros_like(main_results, dtype=np.float32)

        for i, (start, stop) in ProgressBar(
            enumerate(zip(self._starts, self._stops)),
            total=self.nruns,
            desc="Performing run",
        ):

            # needed else trj seek thinks a np.int64 isn't an int?
            results = self._single_run(int(start), int(stop))

            nresults = len(results)
            if nresults == len(main_results):
                main_results += results
                counter += 1.0
            else:
                main_results[:nresults] += results
                counter[:nresults] += 1.0

        main_results /= counter

        self.solution["time"] = (
            np.arange(len(main_results), dtype=np.float32)
            * self.u.trajectory.dt
            * self._skip
        )
        self.solution["results"] = main_results

    def _single_run(self, start, stop):
        """Perform a single pass of the trajectory"""
        self.u.trajectory[start]

        # Calculate partners at t=0
        box = self.u.dimensions if self.pbc else None

        # 2d array of all distances
        pair = capped_distance(
            self.h.positions,
            self.a.positions,
            max_cutoff=self.d_crit,
            box=box,
            return_distances=False,
        )
        if self.exclusions is not None:
            pair = pair[~_in2d(pair, self.exclusions)]

        hidx, aidx = np.transpose(pair)

        a = calc_angles(
            self.d.positions[hidx],
            self.h.positions[hidx],
            self.a.positions[aidx],
            box=box,
        )
        # from amongst those, who also satisfiess angle crit
        idx2 = np.where(a > self.a_crit)
        hidx = hidx[idx2]
        aidx = aidx[idx2]

        nbonds = len(hidx)  # number of hbonds at t=0
        results = np.zeros_like(
            np.arange(start, stop, self._skip), dtype=np.float32
        )

        if self.time_cut:
            # counter for time criteria
            count = np.zeros(nbonds, dtype=np.float64)

        for i, ts in enumerate(self.u.trajectory[start : stop : self._skip]):
            box = self.u.dimensions if self.pbc else None

            d = calc_bonds(
                self.h.positions[hidx], self.a.positions[aidx], box=box
            )
            a = calc_angles(
                self.d.positions[hidx],
                self.h.positions[hidx],
                self.a.positions[aidx],
                box=box,
            )

            winners = (d < self.d_crit) & (a > self.a_crit)
            results[i] = winners.sum()

            if self.bond_type == "continuous":
                # Remove losers for continuous definition
                hidx = hidx[np.where(winners)]
                aidx = aidx[np.where(winners)]
            elif self.bond_type == "intermittent":
                if self.time_cut:
                    # Add to counter of where losers are
                    count[~winners] += self._skip * self.u.trajectory.dt
                    count[winners] = 0  # Reset timer for winners

                    # Remove if you've lost too many times
                    # New arrays contain everything but removals
                    hidx = hidx[count < self.time_cut]
                    aidx = aidx[count < self.time_cut]
                    count = count[count < self.time_cut]
                else:
                    pass

            if len(hidx) == 0:  # Once everyone has lost, the fun stops
                break

        results /= nbonds

        return results

    def solve(self, p_guess=None):
        """Fit results to an multi exponential decay and integrate to find
        characteristic time

        Parameters
        ----------
        p_guess : tuple of floats, optional
            Initial guess for the leastsq fit, must match the shape of the
            expected coefficients


        Continuous defition results are fitted to a double exponential with
        :func:`scipy.optimize.leastsq`, intermittent definition are fit to a
        triple exponential.

        The results of this fitting procedure are saved into the *fit*,
        *tau* and *estimate* keywords in the solution dict.

         - *fit* contains the coefficients, (A1, tau1, tau2) or
           (A1, A2, tau1, tau2, tau3)
         - *tau* contains the calculated lifetime in ps for the hydrogen
           bonding
         - *estimate* contains the estimate provided by the fit of the time
           autocorrelation function

        In addition, the output of the :func:`~scipy.optimize.leastsq` function
        is saved into the solution dict

         - *infodict*
         - *mesg*
         - *ier*

        """

        if self.solution["results"] is None:
            raise ValueError(
                "Results have not been generated use, the run method first"
            )

        # Prevents an odd bug with leastsq where it expects
        # double precision data sometimes...
        time = self.solution["time"].astype(np.float64)
        results = self.solution["results"].astype(np.float64)

        def within_bounds(p):
            """Returns True/False if boundary conditions are met or not.
            Uses length of p to detect whether it's handling continuous /
            intermittent

            Boundary conditions are:
             0 < A_x < 1
             sum(A_x) < 1
             0 < tau_x
            """
            if len(p) == 3:
                A1, tau1, tau2 = p
                return (A1 > 0.0) & (A1 < 1.0) & (tau1 > 0.0) & (tau2 > 0.0)
            elif len(p) == 5:
                A1, A2, tau1, tau2, tau3 = p
                return (
                    (A1 > 0.0)
                    & (A1 < 1.0)
                    & (A2 > 0.0)
                    & (A2 < 1.0)
                    & ((A1 + A2) < 1.0)
                    & (tau1 > 0.0)
                    & (tau2 > 0.0)
                    & (tau3 > 0.0)
                )

        def err(p, x, y):
            """Custom residual function, returns real residual if all
            boundaries are met, else returns a large number to trick the
            leastsq algorithm
            """
            if within_bounds(p):
                return y - self._my_solve(x, *p)
            else:
                return np.full_like(y, 100000)

        def double(x, A1, tau1, tau2):
            """Sum of two exponential functions"""
            A2 = 1 - A1
            return A1 * np.exp(-x / tau1) + A2 * np.exp(-x / tau2)

        def triple(x, A1, A2, tau1, tau2, tau3):
            """Sum of three exponential functions"""
            A3 = 1 - (A1 + A2)
            return (
                A1 * np.exp(-x / tau1)
                + A2 * np.exp(-x / tau2)
                + A3 * np.exp(-x / tau3)
            )

        if self.bond_type == "continuous":
            self._my_solve = double

            if p_guess is None:
                p_guess = (0.5, 10 * self.sample_time, self.sample_time)

                p, cov, infodict, mesg, ier = scipy.optimize.leastsq(
                    err, p_guess, args=(time, results), full_output=True
                )
            self.solution["fit"] = p
            A1, tau1, tau2 = p
            A2 = 1 - A1
            self.solution["tau"] = A1 * tau1 + A2 * tau2
        else:
            self._my_solve = triple

            if p_guess is None:
                p_guess = (
                    0.33,
                    0.33,
                    10 * self.sample_time,
                    self.sample_time,
                    0.1 * self.sample_time,
                )

            p, cov, infodict, mesg, ier = scipy.optimize.leastsq(
                err, p_guess, args=(time, results), full_output=True
            )
            self.solution["fit"] = p
            A1, A2, tau1, tau2, tau3 = p
            A3 = 1 - A1 - A2
            self.solution["tau"] = A1 * tau1 + A2 * tau2 + A3 * tau3

        self.solution["infodict"] = infodict
        self.solution["mesg"] = mesg
        self.solution["ier"] = ier

        if ier in [1, 2, 3, 4]:  # solution found if ier is one of these values
            self.solution["estimate"] = self._my_solve(
                self.solution["time"], *p
            )
        else:
            warnings.warn("Solution to results not found", RuntimeWarning)

    def __repr__(self):
        return (
            "<MDAnalysis HydrogenBondAutoCorrel analysis measuring the "
            "{btype} lifetime of {n} different hydrogens>"
            "".format(btype=self.bond_type, n=len(self.h))
        )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest
import MDAnalysis as mda
import os

from numpy.testing import assert_equal, assert_almost_equal, assert_allclose
import numpy as np

from MDAnalysisTests.coordinates.reference import RefTRZ
from MDAnalysisTests.datafiles import TRZ_psf, TRZ, two_water_gro


def test_deprecated_trz_reader():
    wmsg = "The TRZ reader is deprecated"

    with pytest.warns(DeprecationWarning, match=wmsg):
        _ = mda.Universe(TRZ_psf, TRZ)


def test_deprecated_trz_writer(tmpdir):
    u = mda.Universe(two_water_gro)

    wmsg = "The TRZ writer is deprecated"

    with pytest.warns(DeprecationWarning, match=wmsg):
        with tmpdir.as_cwd():
            with mda.coordinates.TRZ.TRZWriter("test.trz", len(u.atoms)) as W:
                W.write(u)


@pytest.mark.filterwarnings("ignore:The TRZ reader is deprecated")
@pytest.mark.filterwarnings("ignore:The TRZ writer is deprecated")
class TestTRZReader(RefTRZ):
    prec = 3

    @pytest.fixture()
    def universe(self):
        return mda.Universe(TRZ_psf, TRZ)

    def test_load_trz(self, universe):
        U = universe
        assert_equal(
            len(U.atoms), self.ref_n_atoms, "load Universe from PSF and TRZ"
        )

    def test_next_trz(self, universe):
        assert_equal(universe.trajectory.ts.frame, 0, "starts at first frame")
        universe.trajectory.next()
        assert_equal(
            universe.trajectory.ts.frame, 1, "next returns frame index 1"
        )

    def test_rewind_trz(self, universe):
        # move to different frame and rewind to get first frame back
        universe.trajectory[2]
        universe.trajectory.rewind()
        assert_equal(universe.trajectory.ts.frame, 0, "rewinding to frame 1")

    def test_n_frames(self, universe):
        assert_equal(
            universe.trajectory.n_frames,
            self.ref_n_frames,
            "wrong number of frames in trz",
        )

    def test_seeking(self, universe):
        universe.trajectory[3]
        assert_equal(universe.trajectory.ts.frame, 3, "loading frame 3")

        orig = universe.atoms[0:3].positions.copy()

        universe.trajectory[4]
        assert_equal(universe.trajectory.ts.frame, 4, "loading frame 4")
        universe.trajectory[3]

        assert_almost_equal(universe.atoms[0:3].positions, orig, self.prec)

        universe.trajectory[0]
        assert_equal(universe.trajectory.ts.frame, 0, "loading frame 0")
        universe.trajectory[3]

        assert_almost_equal(universe.atoms[0:3].positions, orig, self.prec)

    def test_volume(self, universe):
        # Lower precision here because errors seem to accumulate and
        # throw this off (is rounded value**3)
        assert_almost_equal(
            universe.trajectory.ts.volume,
            self.ref_volume,
            1,
            "wrong volume for trz",
        )

    def test_unitcell(self, universe):
        assert_almost_equal(
            universe.trajectory.ts.dimensions,
            self.ref_dimensions,
            self.prec,
            "wrong dimensions for trz",
        )

    def test_coordinates(self, universe):
        fortytwo = universe.atoms[41]  # 41 because is 0 based
        assert_almost_equal(
            fortytwo.position,
            self.ref_coordinates,
            self.prec,
            "wrong coordinates in trz",
        )

    def test_velocities(self, universe):
        fortytwo = universe.select_atoms("bynum 42")
        assert_almost_equal(
            fortytwo.velocities,
            self.ref_velocities,
            self.prec,
            "wrong velocities in trz",
        )

    def test_delta(self, universe):
        assert_almost_equal(
            universe.trajectory.delta,
            self.ref_delta,
            self.prec,
            "wrong time delta in trz",
        )

    def test_time(self, universe):
        assert_almost_equal(
            universe.trajectory.time,
            self.ref_time,
            self.prec,
            "wrong time value in trz",
        )

    def test_title(self, universe):
        assert_equal(
            self.ref_title, universe.trajectory.title, "wrong title in trz"
        )

    def test_get_writer(self, universe, tmpdir):
        self.outfile = os.path.join(str(tmpdir), "test-trz-writer.trz")
        with universe.trajectory.Writer(self.outfile) as W:
            assert_equal(isinstance(W, mda.coordinates.TRZ.TRZWriter), True)
            assert_equal(W.n_atoms, universe.trajectory.n_atoms)

    def test_get_writer_2(self, universe, tmpdir):
        self.outfile = os.path.join(str(tmpdir), "test-trz-writer-1.trz")
        with universe.trajectory.Writer(self.outfile, n_atoms=100) as W:
            assert_equal(isinstance(W, mda.coordinates.TRZ.TRZWriter), True)
            assert_equal(W.n_atoms, 100)

    def test_get_wrong_n_atoms(self):
        with pytest.raises(ValueError, match=r"Supplied n_atoms"):
            mda.Universe(TRZ, n_atoms=8080)

    def test_read_zero_box(self, tmpdir):
        outfile = str(tmpdir.join("/test-trz-writer.trz"))

        u = mda.Universe.empty(10, trajectory=True)
        u.dimensions = None

        with mda.Writer(outfile, n_atoms=10) as w:
            w.write(u)

        u2 = mda.Universe(outfile, n_atoms=10, to_guess=())

        assert u2.dimensions is None


@pytest.mark.filterwarnings("ignore:The TRZ writer is deprecated")
class TestTRZWriter(RefTRZ):
    prec = 3
    writer = mda.coordinates.TRZ.TRZWriter
    title_to_write = "Test title TRZ"

    @pytest.fixture()
    def universe(self):
        return mda.Universe(TRZ_psf, TRZ)

    @pytest.fixture()
    def outfile(self, tmpdir):
        return str(tmpdir.join("/test-trz-writer.trz"))

    def test_write_trajectory(self, universe, outfile):
        t = universe.trajectory
        W = self.writer(outfile, t.n_atoms, title=self.title_to_write)
        self._copy_traj(W, universe, outfile)

    def _copy_traj(self, writer, universe, outfile):
        for ts in universe.trajectory:
            writer.write(universe)
        writer.close()

        uw = mda.Universe(TRZ_psf, outfile)

        assert_equal(
            uw.trajectory.title,
            self.title_to_write,
            "Title mismatch between original and written files.",
        )

        for orig_ts, written_ts in zip(universe.trajectory, uw.trajectory):
            assert_almost_equal(
                orig_ts._pos,
                written_ts._pos,
                self.prec,
                err_msg="Coordinate mismatch between "
                "orig and written at frame %d" % orig_ts.frame,
            )
            assert_almost_equal(
                orig_ts._velocities,
                written_ts._velocities,
                self.prec,
                err_msg="Coordinate mismatch between "
                "orig and written at frame %d" % orig_ts.frame,
            )
            assert_almost_equal(
                orig_ts._unitcell,
                written_ts._unitcell,
                self.prec,
                err_msg="Unitcell mismatch "
                "between orig and written at frame %d" % orig_ts.frame,
            )
            for att in orig_ts.data:
                assert_almost_equal(
                    orig_ts.data[att],
                    written_ts.data[att],
                    self.prec,
                    err_msg="TS equal failed for {0!s}".format(att),
                )

    def test_long_title(self, outfile):
        title = "*" * 81
        with pytest.raises(ValueError):
            self.writer(outfile, self.ref_n_atoms, title=title)

    def test_no_box_warning(self, outfile):
        u = mda.Universe.empty(10, trajectory=True)
        u.dimensions = None

        with pytest.warns(
            UserWarning, match="box will be written as all zero values"
        ):
            with mda.Writer(outfile, n_atoms=10) as w:
                w.write(u.atoms)


@pytest.mark.filterwarnings("ignore:The TRZ writer is deprecated")
class TestTRZWriter2(object):
    @pytest.fixture()
    def u(self):
        return mda.Universe(two_water_gro)

    @pytest.fixture()
    def outfile(self, tmpdir):
        return str(tmpdir.join("/trz-writer-2.trz"))

    def test_writer_trz_from_other(self, u, outfile):
        with mda.coordinates.TRZ.TRZWriter(outfile, len(u.atoms)) as W:
            W.write(u)

        u2 = mda.Universe(two_water_gro, outfile)

        assert_almost_equal(u.atoms.positions, u2.atoms.positions, 3)

    def test_no_dt_warning(self, u, outfile):
        with mda.coordinates.TRZ.TRZWriter(outfile, len(u.atoms)) as W:
            W.write(u)

        u2 = mda.Universe(two_water_gro, outfile)

        wmsg = "Reader has no dt information, set to 1.0 ps"
        with pytest.warns(UserWarning, match=wmsg):
            assert_allclose(u2.trajectory.dt, 1.0)


@pytest.mark.filterwarnings("ignore:The TRZ writer is deprecated")
class TestWrite_Partial_Timestep(object):
    """Test writing a partial timestep made by passing only an atomgroup to
    Writer. (Issue 163)

    The contents of the AtomGroup.ts are checked in test_atomgroup, this test
    just checks that Writer is receiving this information properly.

    """

    prec = 3

    @pytest.fixture()
    def universe(self):
        return mda.Universe(TRZ_psf, TRZ)

    def test_write_trajectory(self, universe, tmpdir):
        ag = universe.select_atoms("name N")
        outfile = str(tmpdir.join("/partial-write-test.pdb"))
        writer = mda.Writer(outfile, n_atoms=len(ag))
        writer.write(ag)
        writer.close()

        u_ag = mda.Universe(outfile)

        assert_almost_equal(
            ag.positions,
            u_ag.atoms.positions,
            self.prec,
            err_msg="Writing AtomGroup timestep failed.",
        )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
import MDAnalysis as mda
from MDAnalysis.core import topologyattrs
from MDAnalysis.lib.mdamath import triclinic_box
from MDAnalysis.lib.distances import transform_RtoS


def assert_in_box(positions, box):
    """Asserts that all `positions` are strictly within the primary periodic
    image as defined by `box`
    """
    relpos = transform_RtoS(positions, box)
    assert np.all((relpos >= 0.0) & (relpos < 1.0))


class UnWrapUniverse(object):
    """A Universe containing small molecule clusters with molecules (as well as
    clusters) broken or whole accross periodic boundaries:

    The universe comprises 47 atoms in 12 molecules:
    * Three "molecules" (type A) containing a single atom:
     1. within the central image
     2. in a neighboring image
     3. in an image two boxes away from the central image

    * Four molecules (type B) containing three atoms:
     4. within the central image
     5. in the corner of the box, close to 4., broken accross one boundary
        but whole accross another
     6. outside the box, broken diagonally accross two neighboring images,
        close to 5.
     7. whole accross a boundary opposite to the location of 4., i.e.,
        close to 4. in terms of PBC
    * Two cyclical molecules (type C) containing four atoms:
     8. broken accross the front/back but whole accross the top face
     9. within the central image close to the front and bottom face, close
        to 8. in terms of PBC
    * Three linear chain molecules (type D) containing 8 atoms, spanning
     more than half a box length:
     10. within the central image relatively close to the top boundary
     11. close to 10, broken mid-molecule accross the left/right boundary
     12. close to 11. but in another image, whole accross the same
         boundary as 11. but not mid-molecule

    There are 15 residues in the universe:
    Molecules of type A, B, and C each have a single residue, while each of
    the chain molecules of type D have two residues with 4 atoms per residue.

    Atoms can be selected by their residue's resname. For molecules of type
    A, B, and C, the resnames are A, B, and C, respectively. Molecules of
    type D have the resnames D1 and D2.

    Atoms can also be selected by their moltype attribute, which is identical
    to the corresponding molecule type (A, B, C, D).

    The molecules/residues are contained in 6 segments, whereof the first
    three segments contain all molecules of type A, B, or C, respectively.
    Each of the remaining three segments contain a molecule of type D.

    A projection onto the x-z plane of box (orthorhombic case) looks roughly
    like this::

             :                   :
             :                   :
        6    :        8          :                   :
      - + - -:+-------+----------:- - - - - - - - - -: -
             |5       8          |                   :
             |                   |                   :
             |         (10)      |                   :
             |   o-o-o-o-x-x-x-x |                   :
             +x-x-x-x     o-o-o-o+                   :
             |            (11)   |                   :
             |                   |                   :
             |         1         |       2           : 3
             |                   |                   :
             |          9        |                   :
             |   4      !       7|                   :
             |   !      9       !|                   :
             |   4-4            7+7                  :
             |                   |                   :
            5+5                  |                   :
      - + - -:+------------------:- - - - - - - - - -: -
        6-6  :                   :                   :
             :                   :
             :                   :
             :                   :
             :             (12)  :
             :        x-x-x-x-o-o+o-o
             :                   :

    Note that the cyclic structures of molecules 8 and 9 lie in the x-y plane
    and are therefore not fully visible in the x-z plane projection.

    Parameters
    ----------
    have_bonds : bool, optional
        If ``True``, intramolecular bonds will be present in the topology.
        If ``False``, there will be no bond information.
    have_masses : bool, optional
        If ``True``, each atom will be assigned a mass of 1 u.
        If ``False``, masses won't be present in the topology.
    have_molnums : bool, optional
        If ``True``, the topology will have molecule information (molnums).
        If ``False``, that information will be missing.
    have_charges : bool, optional
        If ``False``, charges won't be present in the topology.
        If ``True``, atoms will carry the following charges::
        * atoms of molecule type A: 2 (total: 6)
        * atoms of molecule type B: -0.5 (total: -6)
        * atoms of molecule type C: -1.5 (total: -12)
        * atoms of molecule type D: 0.5 (total: 12)
    is_triclinic : bool, optional
        If ``False``, the box will be a cube with an edge length of 10 Angstrom.
        If ``True``, the cubic box will be sheared by 1 Angstrom in the x and
        y directions.
    """

    def __new__(
        cls,
        have_bonds=True,
        have_masses=True,
        have_molnums=True,
        have_charges=True,
        is_triclinic=False,
    ):
        # box:
        a = 10.0  # edge length
        tfac = 0.1  # factor for box vector shift of triclinic boxes (~84°)
        if is_triclinic:
            box = triclinic_box(
                [a, 0.0, 0.0], [a * tfac, a, 0.0], [a * tfac, a * tfac, a]
            )
        else:
            box = np.array([a, a, a, 90.0, 90.0, 90.0], dtype=np.float32)

        # number of atoms, residues, and segments:
        n_atoms = 47
        n_residues = 15
        n_segments = 6

        # resindices:
        residx = np.empty(n_atoms, dtype=np.int64)
        # type A
        rix = 0
        for i in range(0, 3, 1):
            residx[i] = rix
            rix += 1
        # type B
        for i in range(3, 15, 3):
            residx[i : i + 3] = rix
            rix += 1
        # type C & D
        for i in range(15, 47, 4):
            residx[i : i + 4] = rix
            rix += 1

        # segindices:
        segidx = np.empty(n_residues, dtype=np.int64)
        segidx[0:3] = 0
        segidx[3:7] = 1
        segidx[7:9] = 2
        segidx[9:11] = 3
        segidx[11:13] = 4
        segidx[13:15] = 5

        # universe:
        u = mda.Universe.empty(
            # topology things
            n_atoms=n_atoms,
            n_residues=n_residues,
            n_segments=n_segments,
            atom_resindex=residx,
            residue_segindex=segidx,
            # trajectory things
            trajectory=True,
            velocities=False,
            forces=False,
        )

        # resnames: we always want those for selection purposes
        resnames = ["A"] * 3
        resnames += ["B"] * 4
        resnames += ["C"] * 2
        resnames += ["D1", "D2"] * 3
        u.add_TopologyAttr(topologyattrs.Resnames(resnames))

        # moltypes: we always want those for selection purposes
        moltypes = ["A"] * 3
        moltypes += ["B"] * 4
        moltypes += ["C"] * 2
        moltypes += ["D"] * 6
        u.add_TopologyAttr(topologyattrs.Moltypes(moltypes))

        # trajectory:
        ts = u.trajectory.ts
        ts.frame = 0
        ts.dimensions = box

        # positions:
        relpos = np.empty((n_atoms, 3), dtype=np.float32)
        # type A
        relpos[0:3, :] = np.array(
            [[0.5, 0.5, 0.5], [1.4, 0.5, 0.5], [2.1, 0.5, 0.5]],
            dtype=np.float32,
        )
        # type B
        relpos[3:15, :] = np.array(
            [
                [0.1, 0.1, 0.2],
                [0.1, 0.1, 0.1],
                [0.2, 0.1, 0.1],
                [-0.05, 0.2, 0.05],
                [0.05, 0.2, 0.05],
                [0.05, 0.2, 0.95],
                [-0.2, -0.9, 1.05],
                [-0.2, 0.1, -0.05],
                [-0.1, 0.1, -0.05],
                [0.95, 0.2, 0.25],
                [0.95, 0.2, 0.15],
                [1.05, 0.2, 0.15],
            ],
            dtype=np.float32,
        )
        # type C
        relpos[15:23, :] = np.array(
            [
                [0.4, 0.95, 1.05],
                [0.4, 0.95, 0.95],
                [0.4, 0.05, 0.95],
                [0.4, 0.05, 1.05],
                [0.6, 0.05, 0.25],
                [0.6, 0.05, 0.15],
                [0.6, 0.15, 0.15],
                [0.6, 0.15, 0.25],
            ],
            dtype=np.float32,
        )
        # type D
        relpos[23:47, :] = np.array(
            [
                [0.2, 0.7, 0.8],
                [0.3, 0.7, 0.8],
                [0.4, 0.7, 0.8],
                [0.5, 0.7, 0.8],
                [0.6, 0.7, 0.8],
                [0.7, 0.7, 0.8],
                [0.8, 0.7, 0.8],
                [0.9, 0.7, 0.8],
                [0.66, 0.75, 0.7],
                [0.76, 0.75, 0.7],
                [0.86, 0.75, 0.7],
                [0.96, 0.75, 0.7],
                [0.06, 0.75, 0.7],
                [0.16, 0.75, 0.7],
                [0.26, 0.75, 0.7],
                [0.36, 0.75, 0.7],
                [1.14, 0.65, -0.4],
                [1.04, 0.65, -0.4],
                [0.94, 0.65, -0.4],
                [0.84, 0.65, -0.4],
                [0.74, 0.65, -0.4],
                [0.64, 0.65, -0.4],
                [0.54, 0.65, -0.4],
                [0.44, 0.65, -0.4],
            ],
            dtype=np.float32,
        )
        # make a copy, we need the original later
        _relpos = relpos.copy()
        # apply y- and z-dependent shift of x and y coords for triclinic boxes:
        if is_triclinic:
            # x-coord shift depends on y- and z-coords
            _relpos[:, 0] += tfac * _relpos[:, 1:].sum(axis=1)
            # y-coord shift depends on z-coords only
            _relpos[:, 1] += tfac * _relpos[:, 2]
        # scale relative to absolute positions:
        ts.positions = (_relpos * np.array([a, a, a])).astype(np.float32)

        # bonds:
        if have_bonds:
            bonds = []
            # type A has no bonds
            # type B
            for base in range(3, 15, 3):
                for i in range(2):
                    bonds.append((base + i, base + i + 1))
            # type C
            for base in range(15, 23, 4):
                for i in range(3):
                    bonds.append((base + i, base + i + 1))
                bonds.append((0 + base, 3 + base))
            # type D
            for base in range(23, 47, 8):
                for i in range(7):
                    bonds.append((base + i, base + i + 1))
            u.add_TopologyAttr(topologyattrs.Bonds(bonds))

        # masses:
        if have_masses:
            # masses are all set to 1 so that one can cross-check the results of
            # reference='com' with reference='cog' unwrapping
            masses = np.ones(n_atoms)
            u.add_TopologyAttr(topologyattrs.Masses(masses))

        # molnums:
        if have_molnums:
            molnums = [0, 1, 2]
            molnums += [3, 4, 5, 6]
            molnums += [7, 8]
            molnums += [9, 9, 10, 10, 11, 11]
            u.add_TopologyAttr(topologyattrs.Molnums(molnums))

        # charges:
        if have_charges:
            # type A
            charges = [2] * 3
            # type B
            charges += [-0.5] * 12
            # type C
            charges += [-1.5] * 8
            # type C
            charges += [0.5] * 24
            u.add_TopologyAttr(topologyattrs.Charges(charges))

        # shamelessly monkey-patch some custom universe attributes:
        u._is_triclinic = is_triclinic
        u._relpos = relpos
        u._tfac = tfac
        u._box_edge = a
        # bind custom methods to universe:
        u.unwrapped_coords = cls.unwrapped_coords.__get__(u)
        u.wrapped_coords = cls.wrapped_coords.__get__(u)
        u.center = cls.center.__get__(u)

        return u

    def unwrapped_coords(self, compound, reference):
        """Returns coordinates which correspond to the unwrapped system.

        Parameters
        ----------
        compound : {'group', 'segments', 'residues', 'molecules', 'fragments'}
            Which type of component is unwrapped.
        reference : {'com', 'cog', None}
            The reference point of each compound that is shifted into the
            primary unit cell.

        Note
        ----
        This function assumes that all atom masses are equal. Therefore, the
        returned coordinates for ``reference='com'`` and ``reference='cog'`` are
        identical.
        """
        if reference is not None:
            ref = reference.lower()
            if ref not in ["com", "cog"]:
                raise ValueError(
                    "Unknown unwrap reference: {}" "".format(reference)
                )
        comp = compound.lower()
        if comp not in [
            "group",
            "segments",
            "residues",
            "molecules",
            "fragments",
        ]:
            raise ValueError("Unknown unwrap compound: {}".format(compound))

        # get relative positions:
        relpos = self._relpos.copy()
        # type B
        # molecule 5, atom 2 & molecule 6, atom 1 & 2
        relpos[8, :] = [0.05, 0.2, -0.05]
        relpos[10, :] = [-0.2, -0.9, 0.95]
        relpos[11, :] = [-0.1, -0.9, 0.95]
        # type C
        # molecule 8, atoms 2 & 3
        relpos[17, :] = [0.4, 1.05, 0.95]
        relpos[18, :] = [0.4, 1.05, 1.05]
        # type D
        # molecule 11, residue 1
        relpos[35:39, :] = np.array(
            [
                [1.06, 0.75, 0.7],
                [1.16, 0.75, 0.7],
                [1.26, 0.75, 0.7],
                [1.36, 0.75, 0.7],
            ],
            dtype=np.float32,
        )

        # apply image shifts if necessary:
        if reference is None:
            if comp == "residues":
                # second residue of molecule 11
                relpos[35:39, :] = np.array(
                    [
                        [0.06, 0.75, 0.7],
                        [0.16, 0.75, 0.7],
                        [0.26, 0.75, 0.7],
                        [0.36, 0.75, 0.7],
                    ],
                    dtype=np.float32,
                )
        else:
            # molecule 2 & 3
            relpos[1:3, :] = np.array(
                [[0.4, 0.5, 0.5], [0.1, 0.5, 0.5]], dtype=np.float32
            )
            # molecule 6
            relpos[9:12, :] = np.array(
                [[0.8, 0.1, 1.05], [0.8, 0.1, 0.95], [0.9, 0.1, 0.95]],
                dtype=np.float32,
            )
            # molecule 8
            relpos[15:19, :] = np.array(
                [
                    [0.4, -0.05, 0.05],
                    [0.4, -0.05, -0.05],
                    [0.4, 0.05, -0.05],
                    [0.4, 0.05, 0.05],
                ],
                dtype=np.float32,
            )
            if comp == "residues":
                # molecule 11, residue 1 & molecule 12
                relpos[35:47, :] = np.array(
                    [
                        [0.06, 0.75, 0.7],
                        [0.16, 0.75, 0.7],
                        [0.26, 0.75, 0.7],
                        [0.36, 0.75, 0.7],
                        [1.14, 0.65, 0.6],
                        [1.04, 0.65, 0.6],
                        [0.94, 0.65, 0.6],
                        [0.84, 0.65, 0.6],
                        [0.74, 0.65, 0.6],
                        [0.64, 0.65, 0.6],
                        [0.54, 0.65, 0.6],
                        [0.44, 0.65, 0.6],
                    ],
                    dtype=np.float32,
                )
            else:
                # molecule 11 & 12
                relpos[31:47, :] = np.array(
                    [
                        [-0.34, 0.75, 0.7],
                        [-0.24, 0.75, 0.7],
                        [-0.14, 0.75, 0.7],
                        [-0.04, 0.75, 0.7],
                        [0.06, 0.75, 0.7],
                        [0.16, 0.75, 0.7],
                        [0.26, 0.75, 0.7],
                        [0.36, 0.75, 0.7],
                        [1.14, 0.65, 0.6],
                        [1.04, 0.65, 0.6],
                        [0.94, 0.65, 0.6],
                        [0.84, 0.65, 0.6],
                        [0.74, 0.65, 0.6],
                        [0.64, 0.65, 0.6],
                        [0.54, 0.65, 0.6],
                        [0.44, 0.65, 0.6],
                    ],
                    dtype=np.float32,
                )

        # apply y- and z-dependent shift of x and y coords for triclinic boxes:
        if self._is_triclinic:
            # x-coord shift depends on y- and z-coords
            relpos[:, 0] += self._tfac * relpos[:, 1:].sum(axis=1)
            # y-coord shift depends on z-coords only
            relpos[:, 1] += self._tfac * relpos[:, 2]

        # scale relative to absolute positions:
        a = self._box_edge
        positions = relpos * np.array([a, a, a])

        return positions.astype(np.float32)

    def wrapped_coords(self, compound, center):
        """Returns coordinates which correspond to the wrapped system.

        Parameters
        ----------
        compound : {'atoms', 'group', 'segments', 'residues', 'molecules', \
                    'fragments'}
            Which type of component is unwrapped. Note that for ``'group'``,
            the result will only be correct *if the group is the entire system*.
        center : {'com', 'cog'}
            The reference point of each compound that is shifted into the
            primary unit cell.

        Note
        ----
        This function assumes that all atom masses are equal. Therefore, the
        returned coordinates for ``center='com'`` and ``center='cog'`` are
        identical.
        """
        ctr = center.lower()
        if ctr not in ["com", "cog"]:
            raise ValueError("Unknown unwrap reference: {}".format(center))
        comp = compound.lower()
        if comp not in [
            "atoms",
            "group",
            "segments",
            "residues",
            "molecules",
            "fragments",
        ]:
            raise ValueError("Unknown unwrap compound: {}".format(compound))

        # wrapped relative positions:
        relpos = self._relpos.copy()

        # apply required box shifts:
        if comp == "atoms":
            # type A
            # type A
            # molecule 2: negative x-shift
            relpos[1, 0] -= 1.0
            # molecule 2: negative double x-shift
            relpos[2, 0] -= 2.0
            # type B
            # molecule 5, atom 0: positive x-shift
            relpos[6, 0] += 1.0
            # molecule 6, atom 0: positive x- and y-shift and negative z-shift
            relpos[9, :] += [1.0, 1.0, -1.0]
            # molecule 6, atom 2 & 3: positive x- and z-shift
            relpos[10:12, :] += [1.0, 0.0, 1.0]
            # molecule 7, atom 2: negative x-shift
            relpos[14, 0] -= 1.0
            # type C
            # molecule 8, atoms 0 & 3: negative z-shift
            relpos[15, 2] -= 1.0
            relpos[18, 2] -= 1.0
            # type D
            # molecule 12, atoms 0 & 1: negative x-shift
            relpos[39:41, 0] -= 1.0
            # molecule 12: positive z-shift
            relpos[39:47, 2] += 1.0
        elif comp == "group":
            # com or cog of entire system is within box, so no shift
            pass
        elif comp == "segments":
            # type A
            # molecules 1-3: negative x-shift
            relpos[0:3, 0] -= 1.0
            # type D
            # molecule 12: positive z-shift
            relpos[39:47, 2] += 1.0
        else:  # comp is residues, molecules, or fragments
            # type A
            # molecule 2: negative x-shift
            relpos[1, 0] -= 1.0
            # molecule 2: negative double x-shift
            relpos[2, 0] -= 2.0
            # type B
            # molecule 6: positive x- and y-shift
            relpos[9:12, :2] += 1.0
            # type C
            # molecule 8: negative z-shift
            relpos[15:19, 2] -= 1.0
            # type D
            # molecule 12: positive z-shift
            relpos[39:47, 2] += 1.0

        # apply y- and z-dependent shift of x and y coords for triclinic boxes:
        if self._is_triclinic:
            # x-coord shift depends on y- and z-coords
            relpos[:, 0] += self._tfac * relpos[:, 1:].sum(axis=1)
            # y-coord shift depends on z-coords only
            relpos[:, 1] += self._tfac * relpos[:, 2]

        # scale relative to absolute positions:
        a = self._box_edge
        positions = relpos * np.array([a, a, a])

        return positions.astype(np.float32)

    def center(self, compound):
        """Returns centers which correspond to the unwrapped system.

        Parameters
        ----------
        compound : {'atoms', 'group', 'segments', 'residues', 'molecules', \
                    'fragments'}
            Which type of component is unwrapped. Note that for ``'group'``,
            the result will only be correct *if the group is the entire system*.

        Note
        ----
        This function assumes that all atom masses are equal. Therefore, the
        returned coordinates for ``center='com'`` and ``center='cog'`` are
        identical.
        """

        relpos = self.unwrapped_coords(compound, reference=None)

        comp = compound.lower()
        if comp not in [
            "group",
            "segments",
            "residues",
            "molecules",
            "fragments",
        ]:
            raise ValueError("Unknown unwrap compound: {}".format(compound))

        pos = 0

        if compound == "residues":
            center_pos = np.zeros((15, 3), dtype=np.float32)
        else:
            center_pos = np.zeros((12, 3), dtype=np.float32)

        for base in range(3):
            loc_center = relpos[base, :]
            center_pos[pos, :] = loc_center
            pos += 1

        for base in range(3, 15, 3):
            loc_center = np.mean(relpos[base : base + 3, :], axis=0)
            center_pos[pos, :] = loc_center
            pos += 1

        if compound == "residues":
            for base in range(15, 47, 4):
                loc_center = np.mean(relpos[base : base + 4, :], axis=0)
                center_pos[pos, :] = loc_center
                pos += 1
        else:
            for base in range(15, 23, 4):
                loc_center = np.mean(relpos[base : base + 4, :], axis=0)
                center_pos[pos, :] = loc_center
                pos += 1
            for base in range(23, 47, 8):
                loc_center = np.mean(relpos[base : base + 8, :], axis=0)
                center_pos[pos, :] = loc_center
                pos += 1

        if compound == "group":
            center_pos = center_pos[11]
        elif compound == "segments":
            center_pos = center_pos[9:]

        return center_pos



import numpy as np
import pytest
from numpy.testing import assert_allclose

import MDAnalysis as mda
from MDAnalysis.transformations import NoJump, wrap
from MDAnalysisTests import datafiles as data


@pytest.fixture()
def nojump_universes_fromfile():
    """
    Create the universe objects for the tests.
    """
    u = mda.Universe(data.PSF_TRICLINIC, data.DCD_TRICLINIC)
    transformation = NoJump()
    u.trajectory.add_transformations(transformation)
    return u


@pytest.fixture()
def nojump_universe():
    """
    Create the universe objects for the tests.
    """
    u = mda.Universe.empty(1, trajectory=True)
    coordinates = np.empty((100, u.atoms.n_atoms, 3))  # number of frames
    coordinates[::3, 0] = 0 * np.ones(3) / 3
    coordinates[1::3, 0] = 1 * np.ones(3) / 3
    coordinates[2::3, 0] = 2 * np.ones(3) / 3
    u.load_new(coordinates, order="fac")
    return u


@pytest.fixture()
def nojump_constantvel_universe():
    """
    Create the universe objects for the tests.
    """
    Natom = 1
    Nframe = 100
    coordinates = np.empty((Nframe, Natom, 3))  # number of frames
    coordinates[:, 0, 0] = np.linspace(0, 45, Nframe)
    coordinates[:, 0, 1] = np.linspace(0, 15, Nframe)
    coordinates[:, 0, 2] = np.linspace(0, 10, Nframe)
    reference = mda.Universe.empty(Natom, trajectory=True)
    reference.load_new(coordinates, order="fac")
    return reference


@pytest.fixture
def nojump_universe_npt_2nd_frame():
    """
    Create a Universe in which a single atom jumps across the periodic boundary in the
    x-dimensions at the second frame.

    Unwrapped coordinates should all be 97.5.
    """
    n_atoms = 1
    n_frames = 4
    u = mda.Universe.empty(n_atoms, trajectory=True)
    coordinates = np.empty((n_frames, u.atoms.n_atoms, 3))
    coordinates[0] = [97.5, 50.0, 50.0]
    coordinates[1] = [2.5, 50.0, 50.0]
    coordinates[2] = [2.5, 50.0, 50.0]
    coordinates[3] = [2.5, 50.0, 50.0]
    u.load_new(coordinates, order="fac")
    return u


@pytest.fixture
def nojump_universe_npt_3rd_frame():
    """
    Create a Universe in which a single atom jumps across the periodic boundary in the
    x-dimensions at the third frame.

    Unwrapped coordinates should all be 97.5.
    """
    n_atoms = 1
    n_frames = 4
    u = mda.Universe.empty(n_atoms, trajectory=True)
    coordinates = np.empty((n_frames, u.atoms.n_atoms, 3))
    coordinates[0] = [97.5, 50.0, 50.0]
    coordinates[1] = [97.5, 50.0, 50.0]
    coordinates[2] = [2.5, 50.0, 50.0]
    coordinates[3] = [2.5, 50.0, 50.0]
    u.load_new(coordinates, order="fac")
    return u


@pytest.fixture(scope="module")
def nojump_universe_npt_2nd_frame_from_file(tmp_path_factory):
    """
    Write the `nojump_universe_npt_2nd_frame` fixture to file, read it in and
    return the Universe.

    Used for testing that coordinates can be unwrapped correctly when iterating
    over the trajectory multiple times.

    We can't use an in-memory trajectory to test this because the
    transformation would only be applied once.

    Note, we use `tmp_path_factory` because this fixture requies `module` scope
    so we can read the file after the fixture has been created, and
    `tmp_path` has function-level scope.
    """
    n_atoms = 1
    n_frames = 4
    u = mda.Universe.empty(n_atoms, trajectory=True)
    coordinates = np.empty((n_frames, u.atoms.n_atoms, 3))
    coordinates[0] = [97.5, 50.0, 50.0]
    coordinates[1] = [2.5, 50.0, 50.0]
    coordinates[2] = [2.5, 50.0, 50.0]
    coordinates[3] = [2.5, 50.0, 50.0]
    u.load_new(coordinates, order="fac")
    dim = np.asarray(
        [
            [100, 100, 100, 90, 90, 90],
            [95, 100, 100, 90, 90, 90],  # Box shrinks by 5 in the x-dimension
            [95, 100, 100, 90, 90, 90],
            [95, 100, 100, 90, 90, 90],
        ]
    )
    workflow = [
        mda.transformations.boxdimensions.set_dimensions(dim),
    ]
    u.trajectory.add_transformations(*workflow)
    tmp_pdb = (
        tmp_path_factory.getbasetemp() / "nojump_npt_2nd_frame.pdb"
    ).as_posix()
    tmp_xtc = (
        tmp_path_factory.getbasetemp() / "nojump_npt_2nd_frame.xtc"
    ).as_posix()
    u.atoms.write(tmp_pdb)
    with mda.Writer(tmp_xtc) as f:
        for ts in u.trajectory:
            f.write(u.atoms)
    return mda.Universe(tmp_pdb, tmp_xtc)


def test_nojump_orthogonal_fwd(nojump_universe):
    """
    Test if the nojump transform is returning the correct
    values when iterating forwards over the sample trajectory.
    """
    u = nojump_universe
    dim = np.asarray([1, 1, 1, 90, 90, 90], np.float32)
    workflow = [
        mda.transformations.boxdimensions.set_dimensions(dim),
        NoJump(),
    ]
    u.trajectory.add_transformations(*workflow)
    transformed_coordinates = u.trajectory.timeseries()[0]
    # Step is 1 unit every 3 steps. After 99 steps from the origin,
    # we'll end up at 33.
    assert_allclose(
        transformed_coordinates, np.outer(np.linspace(0, 33, 100), np.ones(3))
    )


def test_nojump_nonorthogonal_fwd(nojump_universe):
    """
    Test if the nojump transform is returning the correct
    values when iterating forwards over the sample trajectory.
    """
    u = nojump_universe
    # Set a non-orthogonal box dimension. The box below works out to be this cell:
    # [[1.        0.        0.       ]
    # [0.        1.        0.       ]
    # [0.5       0.        0.8660254]]
    dim = np.asarray([1, 1, 1, 90, 60, 90], np.float32)
    workflow = [
        mda.transformations.boxdimensions.set_dimensions(dim),
        NoJump(),
    ]
    u.trajectory.add_transformations(*workflow)
    transformed_coordinates = u.trajectory.timeseries()[0]
    # After the transformation, you should end up in a repeating pattern, since you are
    # working in a hexagonal unit cell system. Since you jump every third timestep across
    # a periodic boundary, the shift in each axis is saved. As a consequence, the correct
    # jump every third step is just the original position + the size of the periodic cells.
    assert_allclose(
        transformed_coordinates[::3],
        np.outer(np.arange(33.5), np.array([0.5, 1, np.sqrt(3) / 2])),
    )
    assert_allclose(
        transformed_coordinates[1::3],
        np.outer(np.arange(32.5), np.array([0.5, 1, np.sqrt(3) / 2]))
        + 1 * np.ones(3) / 3,
        rtol=1.2e-7,
    )
    assert_allclose(
        transformed_coordinates[2::3],
        np.outer(np.arange(32.5), np.array([0.5, 1, np.sqrt(3) / 2]))
        + 2 * np.ones(3) / 3,
        rtol=1.2e-7,
    )


def test_nojump_constantvel(nojump_constantvel_universe):
    """
    Test if the nojump transform is returning the correct
    values when iterating forwards over the sample trajectory.
    """
    ref = nojump_constantvel_universe
    towrap = (
        ref.copy()
    )  # This copy of the universe will be wrapped, then unwrapped,
    # and should be equal to ref.
    dim = np.asarray([5, 5, 5, 54, 60, 90], np.float32)
    workflow = [
        mda.transformations.boxdimensions.set_dimensions(dim),
        wrap(towrap.atoms),
        NoJump(),
    ]
    towrap.trajectory.add_transformations(*workflow)
    assert_allclose(
        towrap.trajectory.timeseries(),
        ref.trajectory.timeseries(),
        rtol=5e-07,
        atol=5e-06,
    )


def test_nojump_2nd_frame(nojump_universe_npt_2nd_frame):
    """
    Test if the nojump transform returns the correct values
    at all frames when iterating over an npt trajectory
    and an atom crosses the x-boundary at the second frame

    Wrapped coordinates are:
    coordinates = [
        [97.5, 50.0, 50.0],
        [2.5, 50.0, 50.0],
        [2.5, 50.0, 50.0],
        [2.5, 50.0, 50.0],
    ]
    where each row is a different timestep.

    Unwrapped coordinates are the same at each frame:
    unwrapped = [97.5, 50.0, 50.0]
    """
    u = nojump_universe_npt_2nd_frame
    dim = np.asarray(
        [
            [100, 100, 100, 90, 90, 90],
            [95, 100, 100, 90, 90, 90],  # Box shrinks by 5 in the x-dimension
            [95, 100, 100, 90, 90, 90],
            [95, 100, 100, 90, 90, 90],
        ]
    )
    workflow = [
        mda.transformations.boxdimensions.set_dimensions(dim),
        NoJump(),
    ]
    u.trajectory.add_transformations(*workflow)
    x_position = 97.5
    np.testing.assert_allclose(u.trajectory.timeseries()[:, 0, 0], x_position)


def test_nojump_3rd_frame(nojump_universe_npt_3rd_frame):
    """
    Test if the nojump transform returns the correct values
    at all frames when iterating over an npt trajectory
    and an atom crosses the x-boundary at the third frame.

    Wrapped coordinates are:
    coordinates = [
        [97.5, 50.0, 50.0],
        [97.5, 50.0, 50.0],
        [2.5, 50.0, 50.0],
        [2.5, 50.0, 50.0],
    ]
    where each row is a different timestep.

    Unwarpped coordinates are the same at each frame:
    unwrapped = [97.5, 50.0, 50.0]
    """
    u = nojump_universe_npt_3rd_frame
    dim = np.asarray(
        [
            [100, 100, 100, 90, 90, 90],
            [100, 100, 100, 90, 90, 90],
            [95, 100, 100, 90, 90, 90],  # Box shrinks by 5 in the x-dimension
            [95, 100, 100, 90, 90, 90],
        ]
    )
    workflow = [
        mda.transformations.boxdimensions.set_dimensions(dim),
        NoJump(),
    ]
    u.trajectory.add_transformations(*workflow)
    x_position = 97.5
    np.testing.assert_allclose(u.trajectory.timeseries()[:, 0, 0], x_position)


def test_nojump_iterate_twice(nojump_universe_npt_2nd_frame_from_file):
    """
    Test if the nojump transform always returns the correct values
    at all frames when iterating over multiple times.
    """
    u = nojump_universe_npt_2nd_frame_from_file
    u.trajectory.add_transformations(NoJump())
    timeseries_first_iteration = u.trajectory.timeseries()
    timeseries_second_iteration = u.trajectory.timeseries()
    np.testing.assert_allclose(
        timeseries_first_iteration, timeseries_second_iteration
    )


def test_nojump_constantvel_skip(nojump_universes_fromfile):
    """
    Test if the nojump transform warning is emitted.
    """
    with pytest.warns(UserWarning):
        u = nojump_universes_fromfile
        u.trajectory[0]
        u.trajectory[9]  # Exercises the warning.


def test_nojump_constantvel_stride_2(nojump_universes_fromfile):
    """
    Test if the nojump transform warning is emitted.
    """
    match = "Currently jumping between frames with a step of more than 1."
    with pytest.warns(UserWarning, match=match):
        u = nojump_universes_fromfile
        for ts in u.trajectory[::2]:  # Exercises the warning.
            pass


def test_nojump_constantvel_jumparound(nojump_universes_fromfile):
    """
    Test if the nojump transform is emitting a warning.
    """
    with pytest.warns(UserWarning):
        u = nojump_universes_fromfile
        u.trajectory[0]
        u.trajectory[1]
        u.trajectory[9]


def test_missing_dimensions_init(nojump_universe):
    """
    Test if the nojump transform raises a NoDataError if there is no
    initial dimension for the periodic unit cell.
    """
    with pytest.raises(mda.exceptions.NoDataError):
        u = nojump_universe
        workflow = [NoJump()]
        u.trajectory.add_transformations(*workflow)
        transformed_coordinates = u.trajectory.timeseries()[0]


def test_missing_dimensions(nojump_universe):
    """
    Test if the nojump transform raises a NoDataError if there is no
    dimension for the periodic unit cell in a subsequent timestep.
    """
    with pytest.raises(mda.exceptions.NoDataError):
        u = nojump_universe
        u.dimensions = [73, 73, 73, 90, 90, 90]
        workflow = [NoJump()]
        u.trajectory.add_transformations(*workflow)
        transformed_coordinates = u.trajectory.timeseries()[0]


def test_notinvertible(nojump_universe):
    """
    Test if the nojump transform raises a NoDataError if the dimensions
    are invalid for the periodic unit cell.
    """
    with pytest.raises(mda.exceptions.NoDataError):
        u = nojump_universe
        dim = [1, 0, 0, 90, 90, 90]
        workflow = [
            mda.transformations.boxdimensions.set_dimensions(dim),
            NoJump(),
        ]
        u.trajectory.add_transformations(*workflow)
        transformed_coordinates = u.trajectory.timeseries()[0]
